import json
import requests
import webbrowser
import os
from .endpoints import auth_endpoints_v1, default_global_server_address

default_token_location = "~/.cloudvolume/secrets"
default_token_name = "chunkedgraph-secret.json"
default_token_key = 'token'
default_token_file = f"{default_token_location}/{default_token_name}"


class AuthClient(object):
    """Client to find and use auth tokens to access the dynamic annotation framework services.

    Parameters
    ----------
    token_file : str, optional
        Path to a JSON key:value file holding your auth token.
        By default, "~/.cloudvolume/secrets/chunkedgraph-secret.json"

    token_key : str, optional
        Key for the token in the token_file.
        By default, "token"

    token : str or None, optional
        Direct entry of the token as a string. If provided, overrides the files.
        If None, attempts to use the file paths.

    server_address : str, optional,
        URL to the auth server. By default, uses a default server address.
    """

    def __init__(
        self,
        token_file=None,
        token_key=None,
        token=None,
        server_address=default_global_server_address,
    ):
        if token_file is None:
            token_file = default_token_file
        self._token_file = os.path.expanduser(token_file)

        if token_key is None:
            token_key = default_token_key
        self._token_key = token_key

        if token is None:
            token = self._load_token(self._token_file, self._token_key)
        self._token = token

        self._server_address = server_address
        self._default_endpoint_mapping = {
            "auth_server_address": self._server_address}

    @property
    def token(self):
        """Secret token used to authenticate yourself to the Dynamic Annotation Framework services.
        """
        return self._token

    @token.setter
    def token(self, new_token):
        self._token = new_token
        self._token_key = None

    def get_token(self, token_key=None, ):
        """Load a token with a given key the specified token file

        Parameters
        ----------
        token_key : str or None, optional
            key in the token file JSON, by default None. If None, uses 'token'.
        """
        self._token_key = token_key
        self._token = self._load_token(self._token_file, self._token_key)

    def _load_token(self, token_file, token_key):
        if token_file is None:
            return None

        if os.path.exists(token_file):
            with open(token_file, "r") as f:
                token = json.load(f).get(token_key, None)
        else:
            token = None
        return token

    def get_new_token(self, open=False):
        """Currently, returns instructions for getting a new token based on the current settings and saving it to the local environment. New OAuth tokens are currently not able to be retrieved programmatically.

        Parameters
        ----------
        open : bool, optional
            If True, opens a web browser to the web page where you can generate a new token.
        """
        auth_url = auth_endpoints_v1["refresh_token"].format_map(
            self._default_endpoint_mapping
        )
        txt = f"""New Tokens need to be acquired by hand. Please follow the following steps:
                1) Go to: {auth_url}
                2) Log in with your Google credentials and copy the token shown afterward.
                3a) Save it to your computer with: client.auth.save_token(token="PASTE_YOUR_TOKEN_HERE")
                or
                3b) Set it for the current session only with client.auth.token = "PASTE_YOUR_TOKEN_HERE"
                Note: If you need to save or load multiple tokens, please read the documentation for details.
                Warning! Creating a new token by finishing step 2 will invalidate the previous token!"""
        print(txt)
        if open is True:
            webbrowser.open(auth_url)
        return None

    def save_token(
        self,
        token=None,
        token_key=default_token_key,
        overwrite=False,
        token_file=None,
        switch_token=True,
    ):
        """Conveniently save a token in the correct format.

        After getting a new token by following the instructions in `authclient.get_new_token()`, you can save it with a fully default configuration by running:

        token = 'my_shiny_new_token'

        authclient.save_token(token=token)

        Now on next load, authclient=AuthClient() will make an authclient instance using this token.
        If you would like to specify more information about the json file where the token will be stored, see the parameters below.

        Parameters
        ----------
        token : str, optional
            New token to save, by default None
        token_key : str, optional
            Key for the token in the token_file json, by default "token"
        overwrite : bool, optional
            Allow an existing token to be changed, by default False
        token_file : str, optional
            Path to the token file, by default None. If None, uses the default file location specified above.
        switch_token : bool, optional
            If True, switch the auth client over into using the new token, by default True
        """
        if token is None:
            token = self.token

        if token_file is not None:
            save_token_file = token_file
        else:
            save_token_file = self._token_file

        if save_token_file is None:
            raise ValueError("No token file is set")

        if os.path.exists(save_token_file):
            with open(save_token_file, "r") as f:
                secrets = json.load(f)

            if overwrite is False and token_key in secrets:
                raise ValueError(
                    f"Key \"{token_key}\" already exists in token file \"{save_token_file}\"")
        else:
            secrets = {}

        secrets[token_key] = token

        secret_dir, _ = os.path.split(save_token_file)
        if not os.path.exists(secret_dir):
            full_dir = os.path.expanduser(secret_dir)
            os.makedirs(full_dir)

        with open(save_token_file, "w") as f:
            json.dump(secrets, f)

        if switch_token:
            self._token = token
            self._token_key = token_key
            self._token_file = save_token_file

    @property
    def request_header(self):
        """Formatted request header with the specified token
        """
        if self.token is not None:
            auth_header = {"Authorization": f"Bearer {self.token}"}
            return auth_header
        else:
            return {}



import requests
import json


class AuthException(Exception):
    pass


def handle_response(response, as_json=True):
    '''Deal with potential errors in endpoint response and return json for default case'''
    response.raise_for_status()
    _check_authorization_redirect(response)
    if as_json:
        return response.json()
    else:
        return response


def _check_authorization_redirect(response):
    if len(response.history) == 0:
        pass
    else:
        raise AuthException(
            f"""You do not have permission to use the endpoint {response.history[0].url} with the current auth configuration.\nRead the documentation or follow instructions under client.auth.get_new_token() for how to set a valid API token.""")


def _api_versions(server_name, server_address, endpoints_common, auth_header):
    """Asks a server what API versions are available, if possible
    """
    url_mapping = {server_name: server_address}
    url_base = endpoints_common.get('get_api_versions', None)
    if url_base is not None:
        url = url_base.format_map(url_mapping)
        response = requests.get(url, headers=auth_header)
        self.raise_for_status(response)
        return response.json()
    else:
        return None


def _api_endpoints(api_version, server_name, server_address, endpoints_common, endpoint_versions, auth_header):
    "Gets the latest client API version"
    if api_version == 'latest':
        try:
            avail_vs_server = _api_versions(server_name,
                                            server_address,
                                            endpoints_common,
                                            auth_header
                                            )
            avail_vs_server = set(avail_vs_server)
        except:
            avail_vs_server = None

        avail_vs_client = set(endpoint_versions.keys())

        if avail_vs_server is None:
            api_version = max(avail_vs_client)
        else:
            api_version = max(avail_vs_client.intersection(avail_vs_server))

    endpoints = endpoints_common.copy()
    ep_to_add = endpoint_versions.get(api_version, None)
    if ep_to_add is None:
        raise ValueError('No corresponding API version')
    endpoints.update(ep_to_add)
    return endpoints, api_version


class ClientBase(object):
    def __init__(self,
                 server_address,
                 auth_header,
                 api_version,
                 endpoints,
                 server_name,
                 ):
        self._server_address = server_address
        self._default_url_mapping = {server_name: self._server_address}

        self.session = requests.Session()
        head_val = auth_header.get('Authorization', None)
        if head_val is not None:
            token = head_val.split(' ')[1]
            cookie_obj = requests.cookies.create_cookie(name='middle_auth_token',
                                                        value=token)
            self.session.cookies.set_cookie(cookie_obj)
        self.session.headers.update(auth_header)

        self._api_version = api_version
        self._endpoints = endpoints

    @property
    def default_url_mapping(self):
        return self._default_url_mapping

    @property
    def server_address(self):
        return self._server_address

    @property
    def api_version(self):
        return self._api_version
    
    @staticmethod
    def raise_for_status(r):
        """Raises :class:`HTTPError`, if one occurred."""

        http_error_msg = ''
        if isinstance(r.reason, bytes):
            # We attempt to decode utf-8 first because some servers
            # choose to localize their reason strings. If the string
            # isn't utf-8, we fall back to iso-8859-1 for all other
            # encodings. (See PR #3538)
            try:
                reason = r.reason.decode('utf-8')
            except UnicodeDecodeError:
                reason = r.reason.decode('iso-8859-1')
        else:
            reason = r.reason

        if 400 <= r.status_code < 500:
            http_error_msg = u'%s Client Error: %s for url: %s content: %s' % (r.status_code, reason, r.url, r.content)

        elif 500 <= r.status_code < 600:
            http_error_msg = u'%s Server Error: %s for url: %s content:%s' % (r.status_code, reason, r.url, r.content)

        if http_error_msg:
            raise requests.HTTPError(http_error_msg, response=r)


class ClientBaseWithDataset(ClientBase):
    def __init__(self,
                 server_address,
                 auth_header,
                 api_version,
                 endpoints,
                 server_name,
                 dataset_name
                 ):

        super(ClientBaseWithDataset, self).__init__(server_address,
                                                    auth_header,
                                                    api_version,
                                                    endpoints,
                                                    server_name,
                                                    )
        self._dataset_name = dataset_name

    @property
    def dataset_name(self):
        return self._dataset_name


class ClientBaseWithDatastack(ClientBase):
    def __init__(self,
                 server_address,
                 auth_header,
                 api_version,
                 endpoints,
                 server_name,
                 datastack_name
                 ):

        super(ClientBaseWithDatastack, self).__init__(server_address,
                                                      auth_header,
                                                      api_version,
                                                      endpoints,
                                                      server_name,
                                                      )
        self._datastack_name = datastack_name

    @property
    def datastack_name(self):
        return self._datastack_name



from .base import ClientBase, _api_endpoints, handle_response
from .endpoints import schema_common, schema_api_versions, schema_endpoints_common
from .auth import AuthClient
import requests

server_key = "emas_server_address"


def SchemaClient(server_address=None,
                 auth_client=None,
                 api_version='latest'):
    if auth_client is None:
        auth_client = AuthClient()

    auth_header = auth_client.request_header
    endpoints, api_version = _api_endpoints(api_version, server_key, server_address,
                                            schema_endpoints_common, schema_api_versions, auth_header)
    SchemaClient = client_mapping[api_version]
    return SchemaClient(server_address=server_address,
                        auth_header=auth_header,
                        api_version=api_version,
                        endpoints=endpoints,
                        server_name=server_key)


class SchemaClientLegacy(ClientBase):
    def __init__(self, server_address, auth_header, api_version, endpoints, server_name):
        super(SchemaClientLegacy, self).__init__(server_address,
                                                 auth_header, api_version, endpoints, server_name)

    def get_schemas(self):
        """Get the available schema types

        Returns
        -------
        list
            List of schema types available on the Schema service.
        """
        endpoint_mapping = self.default_url_mapping
        url = self._endpoints['schema'].format_map(endpoint_mapping)
        response = self.session.get(url)
        return handle_response(response)

    def schema_definition(self, schema_type):
        """Get the definition of a specified schema_type

        Parameters
        ----------
        schema_type : str
            Name of a schema_type

        Returns
        -------
        json
            Schema definition
        """
        endpoint_mapping = self.default_url_mapping
        endpoint_mapping['schema_type'] = schema_type
        url = self._endpoints['schema_definition'].format_map(endpoint_mapping)
        response = self.session.get(url)
        return handle_response(response)


client_mapping = {1: SchemaClientLegacy,
                  2: SchemaClientLegacy,
                  'latest': SchemaClientLegacy}


import numpy as np
import requests
import datetime
import time
import json
from . import endpoints
from . import infoservice
from .endpoints import chunkedgraph_api_versions, chunkedgraph_endpoints_common, default_global_server_address
from .base import _api_endpoints, _api_versions, ClientBase, handle_response
from .auth import AuthClient

SERVER_KEY = 'cg_server_address'


class CGEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        if isinstance(obj, np.uint64):
            return int(obj)
        if isinstance(obj, (datetime.datetime, datetime.date)):
            return obj.isoformat()
        return json.JSONEncoder.default(self, obj)


def package_bounds(bounds):
    if (bounds.shape != (3,2)):
        raise ValueError('Bounds must be a 3x2 matrix (x,y,z) x (min,max) in chunkedgraph resolution voxel units')
    
    bounds_str = []
    for b in bounds:
        bounds_str.append("-".join(str(b2) for b2 in b))
    bounds_str = "_".join(bounds_str)
    return bounds_str


def package_timestamp(timestamp):
    if timestamp is None:
        query_d = None
    else:
        query_d = {'timestamp': time.mktime(timestamp.timetuple())}
    return query_d


def ChunkedGraphClient(server_address=None,
                       table_name=None,
                       auth_client=None,
                       api_version='latest',
                       timestamp=None
                       ):
    if server_address is None:
        server_address = default_global_server_address

    if auth_client is None:
        auth_client = AuthClient()

    auth_header = auth_client.request_header

    endpoints, api_version = _api_endpoints(api_version, SERVER_KEY, server_address,
                                            chunkedgraph_endpoints_common, chunkedgraph_api_versions, auth_header)

    ChunkedClient = client_mapping[api_version]
    return ChunkedClient(server_address,
                         auth_header,
                         api_version,
                         endpoints,
                         SERVER_KEY,
                         timestamp=timestamp,
                         table_name=table_name)


class ChunkedGraphClientV1(ClientBase):
    """ ChunkedGraph Client for the v1 API
    """

    def __init__(self, server_address, auth_header, api_version, endpoints, server_key=SERVER_KEY, timestamp=None, table_name=None):
        super(ChunkedGraphClientV1, self).__init__(server_address,
                                                   auth_header,
                                                   api_version,
                                                   endpoints,
                                                   server_key)
        self._default_url_mapping['table_id'] = table_name
        self._default_timestamp = timestamp
        self._table_name = table_name
        self._segmentation_info = None

    @property
    def default_url_mapping(self):
        return self._default_url_mapping.copy()

    @property
    def table_name(self):
        return self._table_name

    def _process_timestamp(self, timestamp):
        """Process timestamp with default logic
        """
        if timestamp is None:
            if self._default_timestamp is not None:
                return self._default_timestamp
            else:
                return datetime.datetime.utcnow()
        else:
            return timestamp

    def get_roots(self, supervoxel_ids, timestamp=None, stop_level=None):
        """Get the root id for a specified supervoxel

        Parameters
        ----------
        supervoxel_ids : np.array(np.uint64)
            Supervoxel ids values
        timestamp : datetime.datetime, optional
            UTC datetime to specify the state of the chunkedgraph at which to query, by default None. If None, uses the current time.
        stop_level : int or None, optional
            If True, looks up ids only up to a given stop level. Default is None.

        Returns
        -------
        np.array(np.uint64)
            Root IDs containing each supervoxel.
        """
        if stop_level is not None:
            return self._get_level_roots(supervoxel_ids, stop_level=stop_level, timestamp=timestamp)

        endpoint_mapping = self.default_url_mapping
        url = self._endpoints['get_roots'].format_map(endpoint_mapping)
        query_d = package_timestamp(self._process_timestamp(timestamp))
        data = np.array(supervoxel_ids, dtype=np.uint64).tobytes()

        response = self.session.post(url, data=data, params=query_d)
        handle_response(response, as_json=False)
        return np.frombuffer(response.content, dtype=np.uint64)

    def get_root_id(self, supervoxel_id, timestamp=None, level2=False):
        """Get the root id for a specified supervoxel

        Parameters
        ----------
        supervoxel_id : np.uint64
            Supervoxel id value
        timestamp : datetime.datetime, optional
            UTC datetime to specify the state of the chunkedgraph at which to query, by default None. If None, uses the current time.

        Returns
        -------
        np.uint64
            Root ID containing the supervoxel.
        """
        endpoint_mapping = self.default_url_mapping
        endpoint_mapping['supervoxel_id'] = supervoxel_id

        url = self._endpoints['handle_root'].format_map(endpoint_mapping)
        query_d = package_timestamp(self._process_timestamp(timestamp))
        if level2:
            query_d['stop_layer'] = 2

        response = self.session.get(url, params=query_d)
        handle_response(response, as_json=False)
        return np.int64(response.json()['root_id'])

    def _get_level_roots(self, supervoxel_ids, stop_level, timestamp=None):
        """Get the root id for a specified supervoxel

        Parameters
        ----------
        supervoxel_id : np.uint64
            Supervoxel id value
        timestamp : datetime.datetime, optional
            UTC datetime to specify the state of the chunkedgraph at which to query, by default None. If None, uses the current time.

        Returns
        -------
        np.uint64
            Root ID containing the supervoxel.
        """
        endpoint_mapping = self.default_url_mapping

        url = self._endpoints['handle_roots'].format_map(endpoint_mapping)
        query_d = package_timestamp(self._process_timestamp(timestamp))
        query_d['stop_layer'] = stop_level
        data = {'node_ids': supervoxel_ids}
        response = self.session.post(
            url, data=json.dumps(data, cls=CGEncoder), params=query_d)
        return handle_response(response, as_json=True).get('root_ids', None)

    def get_merge_log(self, root_id):
        """Get the merge log (splits and merges) for an object

        Parameters
        ----------
        root_id : np.uint64
            Object root id to look up

        Returns
        -------
        list
            List of merge events in the history of the object.
        """
        endpoint_mapping = self.default_url_mapping
        endpoint_mapping['root_id'] = root_id
        url = self._endpoints['merge_log'].format_map(endpoint_mapping)

        response = self.session.get(url)
        return handle_response(response)

    def get_change_log(self, root_id):
        """Get the change log (splits and merges) for an object

        Parameters
        ----------
        root_id : np.uint64
            Object root id to look up

        Returns
        -------
        list
            List of split and merge events in the object history
        """
        endpoint_mapping = self.default_url_mapping
        endpoint_mapping['root_id'] = root_id
        url = self._endpoints['change_log'].format_map(endpoint_mapping)
        response = self.session.get(url)
        return handle_response(response)

    def get_leaves(self, root_id, bounds=None):
        """Get all supervoxels for a root_id

        Parameters
        ----------
        root_id : np.uint64
            Root id to query
        bounds: np.array or None, optional
            If specified, returns supervoxels within a 3x2 numpy array of bounds [[minx,maxx],[miny,maxy],[minz,maxz]]
            If None, finds all supervoxels.

        Returns
        -------
        list
            List of supervoxel ids
        """
        endpoint_mapping = self.default_url_mapping
        endpoint_mapping['root_id'] = root_id
        url = self._endpoints['leaves_from_root'].format_map(endpoint_mapping)
        query_d = {}
        if bounds is not None:
            query_d['bounds'] = package_bounds(bounds)

        response = self.session.get(url, params=query_d)
        handle_response(response, as_json=False)
        return np.int64(response.json()['leaf_ids'])

    def do_merge(self, supervoxels, coords, resolution=(4, 4, 40)):
        """Perform a merge on the chunkeded graph

        Args:
            supervoxels (iterable): a N long list of supervoxels to merge
            coords (np.array): a Nx3 array of coordinates of the supervoxels in units of resolution
            resolution (tuple, optional): what to multiple the coords by to get nm. Defaults to (4,4,40).
        """
        endpoint_mapping = self.default_url_mapping
        url = self._endpoints['do_merge'].format_map(endpoint_mapping)

        data = []
        for svid, coor in zip(supervoxels, coords):
            row = np.concatenate([[svid], np.array(coor)*resolution])
            data.append(row)
        params = {"priority": False}
        response = self.session.post(url, data=json.dumps(data, cls=CGEncoder),
                                     params=params,
                                     headers={'Content-Type': 'application/json'})
        handle_response(response)

    def get_children(self, node_id):
        """Get the children of a node in the hierarchy

        Parameters
        ----------
        node_id : np.uint64
            Node id to query

        Returns
        -------
        list
            List of np.uint64 ids of child nodes.
        """
        endpoint_mapping = self.default_url_mapping
        endpoint_mapping['node_id'] = node_id
        url = self._endpoints['handle_children'].format_map(endpoint_mapping)

        response = self.session.post(url)
        handle_response(response, as_json=False)
        return np.frombuffer(response.content, dtype=np.uint64)

    def get_contact_sites(self, root_id, bounds, calc_partners=False):
        """Get contacts for a root id

        Parameters
        ----------
        root_id : np.uint64
            Object root id
        bounds: np.array
            Bounds within a 3x2 numpy array of bounds [[minx,maxx],[miny,maxy],[minz,maxz]] for which to find contacts. Running this query without bounds is too slow.
        calc_partners : bool, optional
            If True, get partner root ids. By default, False.
        Returns
        -------
        dict
            Dict relating ids to contacts
        """
        endpoint_mapping = self.default_url_mapping
        endpoint_mapping['root_id'] = root_id
        url = self._endpoints['contact_sites'].format_map(endpoint_mapping)
        query_d = {}
        if bounds is not None:
            query_d['bounds'] = package_bounds(bounds)
        query_d['partners'] = calc_partners
        response = self.session.post(url, json=[root_id], params=query_d)
        contact_d = handle_response(response)
        return {int(k): v for k, v in contact_d.items()}

    def find_path(self, root_id, src_pt, dst_pt, precision_mode=False):
        """find a path between two locations on a root_id using the supervoxel lvl2 graph.

        Args:
            root_id (np.int64): the root id to search on
            src_pt (np.array): len(3) xyz location of the start location in nm
            dst_pt ([type]): len(3) xyz location of the end location in nm
            precision_mode (bool, optional): Whether to perform the search in precision mode. Defaults to False.

        Returns:
            centroids_list: centroids
            l2_path: l2_path
            failed_l2_ids: failed_l2_ids
        """
        endpoint_mapping = self.default_url_mapping
        endpoint_mapping['root_id'] = root_id
        url = self._endpoints['find_path'].format_map(endpoint_mapping)
        query_d = {}
        query_d['precision_mode'] = precision_mode

        nodes = [[root_id] + src_pt.tolist(),
                 [root_id] + dst_pt.tolist()]

        response = self.session.post(url,
                                     data=json.dumps(nodes, cls=CGEncoder),
                                     params=query_d,
                                     headers={'Content-Type': 'application/json'})
        resp_d = handle_response(response)
        centroids = np.array(resp_d['centroids_list'])
        failed_l2_ids = np.array(resp_d['failed_l2_ids'], dtype=np.uint64)
        l2_path = np.array(resp_d['l2_path'])

        return centroids, l2_path, failed_l2_ids
    
    def get_subgraph(self, root_id, bounds):
        """Get subgraph of root id within a bounding box

        Args:
            root_id ([int64]): root (or seg_id/node_id) of chunkedgraph to query
            bounds ([np.array]): 3x2 bounding box (x,y,z)x (min,max) in chunkedgraph coordinates
        """
        endpoint_mapping = self.default_url_mapping
        endpoint_mapping['root_id'] = root_id
        url = self._endpoints['get_subgraph'].format_map(endpoint_mapping)
        query_d = {}
        if bounds is not None:
            query_d['bounds'] = package_bounds(bounds)
        
        response = self.session.get(url, params=query_d)
        rd = handle_response(response)
        return np.int64(rd['nodes']), np.double(rd['affinities']), np.int32(rd['areas'])

    def level2_chunk_graph(self, root_id):
        """Get graph of level 2 chunks, the smallest agglomeration level above supervoxels.

        Parameters
        ----------
        root_id : int
            Root id of object

        Returns
        -------
        edge_list : list
            Edge array of level 2 ids
        """
        endpoint_mapping = self.default_url_mapping
        endpoint_mapping['root_id'] = root_id
        url = self._endpoints['lvl2_graph'].format_map(endpoint_mapping)
        r = handle_response(self.session.get(url))
        return r['edge_graph']

    def remesh_level2_chunks(self, chunk_ids):
        """Submit specific level 2 chunks to be remeshed in case of a problem.

        Parameters
        ----------
        chunk_ids : list
            List of level 2 chunk ids.
        """

        endpoint_mapping = self.default_url_mapping
        url = self._endpoints['remesh_level2_chunks'].format_map(
            endpoint_mapping)
        data = {'new_lvl2_ids': [int(x) for x in chunk_ids]}
        r = self.session.post(url, json=data)
        r.raise_for_status()

    @property
    def cloudvolume_path(self):
        return self._endpoints['cloudvolume_path'].format_map(self.default_url_mapping)

    @property
    def segmentation_info(self):
        """Complete segmentation metadata
        """
        if self._segmentation_info is None:
            url = self._endpoints['info'].format_map(self.default_url_mapping)
            response = self.session.get(url)
            self._segmentation_info = handle_response(response)
        return self._segmentation_info

    @property
    def base_resolution(self):
        """MIP 0 resolution for voxels assumed by the ChunkedGraph

        Returns
        -------
        list
            3-long list of x/y/z voxel dimensions in nm
        """
        return self.segmentation_info['scales'][0].get('resolution')


class ChunkedGraphClientLegacy(ClientBase):
    """Client to interface with the PyChunkedGraph service

    Parameters
    ----------
    server_address : str or None, optional
        URL where the PyChunkedGraph service is running. If None, defaults to www.dynamicannotationframework.com
    table_name : str or None, optional
        Name of the chunkedgraph table associated with the dataset. If the datastack_name is specified and table name is not, this can be looked up automatically. By default, None.
    auth_client : auth.AuthClient or None, optional
        Instance of an AuthClient with token to handle authorization. If None, does not specify a token.
    timestamp : datetime.datetime or None, optional
        Default UTC timestamp to use for chunkedgraph queries.
    """

    def __init__(self,
                 server_address,
                 auth_header,
                 api_version,
                 endpoints,
                 server_key=SERVER_KEY,
                 timestamp=None,
                 table_name=None):
        super(ChunkedGraphClientLegacy, self).__init__(server_address,
                                                       auth_header,
                                                       api_version,
                                                       endpoints,
                                                       server_key)

        self._default_url_mapping['table_id'] = table_name
        self._default_timestamp = timestamp
        self._table_name = table_name

    @ property
    def default_url_mapping(self):
        return self._default_url_mapping.copy()

    @ property
    def table_name(self):
        return self._table_name

    def get_roots(self, supervoxel_ids, timestamp=None):
        """Get the root id for a specified supervoxel

        Parameters
        ----------
        supervoxel_ids : np.array(np.uint64)
            Supervoxel ids values
        timestamp : datetime.datetime, optional
            UTC datetime to specify the state of the chunkedgraph at which to query, by default None. If None, uses the current time.

        Returns
        -------
        np.array(np.uint64)
            Root IDs containing each supervoxel.
        """
        if timestamp is None:
            if self._default_timestamp is not None:
                timestamp = self._default_timestamp
            else:
                timestamp = datetime.datetime.utcnow()

        endpoint_mapping = self.default_url_mapping
        url = self._endpoints['get_roots'].format_map(endpoint_mapping)

        if timestamp is None:
            timestamp = self._default_timestamp
        if timestamp is not None:
            query_d = {
                'timestamp': time.mktime(timestamp.timetuple())
            }
        else:
            query_d = None
        data = np.array(supervoxel_ids, dtype=np.uint64).tobytes()

        response = self.session.post(url, data=data, params=query_d)
        handle_response(response, as_json=False)
        return np.frombuffer(response.content, dtype=np.uint64)

    def get_root_id(self, supervoxel_id, timestamp=None, level2=False):
        """Get the root id for a specified supervoxel

        Parameters
        ----------
        supervoxel_id : np.uint64
            Supervoxel id value
        timestamp : datetime.datetime, optional
            UTC datetime to specify the state of the chunkedgraph at which to query, by default None. If None, uses the current time.
        level2 : bool, optional
            Return true to get the level 2 id of a supervoxel.

        Returns
        -------
        np.uint64
            Root ID containing the supervoxel (or level 2 id if requested)
        """
        if timestamp is None:
            if self._default_timestamp is not None:
                timestamp = self._default_timestamp
            else:
                timestamp = datetime.datetime.utcnow()

        endpoint_mapping = self.default_url_mapping
        endpoint_mapping['supervoxel_id'] = supervoxel_id
        url = self._endpoints['handle_root'].format_map(endpoint_mapping)

        if timestamp is None:
            timestamp = self._default_timestamp

        query_d = {}
        if timestamp is not None:
            query_d['timestamp'] = time.mktime(timestamp.timetuple())
        if level2:
            query_d['stop_layer'] = 2

        response = self.session.get(url, params=query_d)
        return np.int64(handle_response(response)['root_id'])

    def get_merge_log(self, root_id):
        """Returns the merge log for a given object

        Parameters
        ----------
        root_id : np.uint64
            Root id of an object to get merge information.

        Returns
        -------
        list
            List of merge events in the history of the object.
        """
        endpoint_mapping = self.default_url_mapping
        endpoint_mapping['root_id'] = root_id
        url = self._endpoints['merge_log'].format_map(endpoint_mapping)
        response = self.session.post(url, json=[root_id])
        return handle_response(response)

    def get_change_log(self, root_id):
        """Get the change log (splits and merges) for an object

        Parameters
        ----------
        root_id : np.uint64
            Object root id to look up

        Returns
        -------
        list
            List of split and merge events in the object history
        """
        endpoint_mapping = self.default_url_mapping
        endpoint_mapping['root_id'] = root_id
        url = self._endpoints['change_log'].format_map(endpoint_mapping)
        response = self.session.post(url, json=[root_id])
        return handle_response(response)

    def get_leaves(self, root_id, bounds=None):
        """Get all supervoxels for a root_id

        Parameters
        ----------
        root_id : np.uint64
            Root id to query
        bounds: np.array or None, optional
            If specified, returns supervoxels within a 3x2 numpy array of bounds [[minx,maxx],[miny,maxy],[minz,maxz]]
            If None, finds all supervoxels.

        Returns
        -------
        list
            List of supervoxel ids
        """
        endpoint_mapping = self.default_url_mapping
        endpoint_mapping['root_id'] = root_id
        url = self._endpoints['leaves_from_root'].format_map(endpoint_mapping)
        query_d = {}
        if bounds is not None:
            query_d['bounds'] = package_bounds(bounds)

        response = self.session.get(url, params=query_d)
        return np.int64(handle_response(response)['leaf_ids'])

    def do_merge(self, supervoxels, coords, resolution=(4, 4, 40)):
        """Perform a merge on the chunkeded graph

        Args:
            supervoxels (iterable): a N long list of supervoxels to merge
            coords (np.array): a Nx3 array of coordinates of the supervoxels in units of resolution
            resolution (tuple, optional): what to multiple the coords by to get nm. Defaults to (4,4,40).
        """
        endpoint_mapping = self.default_url_mapping
        url = self._endpoints['do_merge'].format_map(endpoint_mapping)

        data = []
        for svid, coor in zip(supervoxels, coords):
            row = np.concatenate([[svid], np.array(coor)*resolution])
            data.append(row)
        params = {"priority": False}
        response = self.session.post(url, data=json.dumps(data, cls=CGEncoder),
                                     params=params,
                                     headers={'Content-Type': 'application/json'})
        return handle_response(response)

    def get_children(self, node_id):
        """Get the children of a node in the hierarchy

        Parameters
        ----------
        node_id : np.uint64
            Node id to query

        Returns
        -------
        list
            List of np.uint64 ids of child nodes.
        """
        endpoint_mapping = self.default_url_mapping
        endpoint_mapping['node_id'] = node_id
        url = self._endpoints['handle_children'].format_map(endpoint_mapping)

        response = self.session.post(url)
        handle_response(response, as_json=False)
        return np.frombuffer(response.content, dtype=np.uint64)

    def get_contact_sites(self, root_id, bounds, calc_partners=False):
        """Get contacts for a root id

        Parameters
        ----------
        root_id : np.uint64
            Object root id
        bounds: np.array
            Bounds within a 3x2 numpy array of bounds [[minx,maxx],[miny,maxy],[minz,maxz]] for which to find contacts. Running this query without bounds is too slow.
        calc_partners : bool, optional
            If True, get partner root ids. By default, False.
        Returns
        -------
        dict
            Dict relating ids to contacts
        """
        endpoint_mapping = self.default_url_mapping
        endpoint_mapping['root_id'] = root_id
        url = self._endpoints['contact_sites'].format_map(endpoint_mapping)
        query_d = {}
        if bounds is not None:
            query_d['bounds'] = package_bounds(bounds)
        query_d['partners'] = calc_partners
        response = self.session.post(url, json=[root_id], params=query_d)
        contact_d = handle_response(response)
        return {int(k): v for k, v in contact_d.items()}

    def find_path(self, root_id, src_pt, dst_pt, precision_mode=False):
        """find a path between two locations on a root_id using the supervoxel lvl2 graph.

        Args:
            root_id (np.int64): the root id to search on
            src_pt (np.array): len(3) xyz location of the start location in nm
            dst_pt ([type]): len(3) xyz location of the end location in nm
            precision_mode (bool, optional): Whether to perform the search in precision mode. Defaults to False.

        Returns:
            centroids_list: centroids
            l2_path: l2_path
            failed_l2_ids: failed_l2_ids
        """
        endpoint_mapping = self.default_url_mapping
        endpoint_mapping['root_id'] = root_id
        url = self._endpoints['find_path'].format_map(endpoint_mapping)
        query_d = {}
        query_d['precision_mode'] = precision_mode

        nodes = [[root_id] + src_pt.tolist(),
                 [root_id] + dst_pt.tolist()]

        response = self.session.post(url,
                                     data=json.dumps(nodes, cls=CGEncoder),
                                     params=query_d,
                                     headers={'Content-Type': 'application/json'})
        resp_d = handle_response(response)
        centroids = np.array(resp_d['centroids_list'])
        failed_l2_ids = np.array(resp_d['failed_l2_ids'], dtype=np.uint64)
        l2_path = np.array(resp_d['l2_path'])

        return centroids, l2_path, failed_l2_ids


client_mapping = {0: ChunkedGraphClientLegacy,
                  1: ChunkedGraphClientV1,
                  'latest': ChunkedGraphClientV1}



from .base import ClientBase, _api_versions, _api_endpoints, handle_response
from .auth import AuthClient
from .endpoints import jsonservice_common, jsonservice_api_versions, default_global_server_address
import requests
import json
import re

server_key = 'json_server_address'


def JSONService(server_address=None,
                auth_client=None,
                api_version='latest',
                ngl_url=None,
                ):
    """Client factory to interface with the JSON state service.

    Parameters
    ----------
    server_address : str, optional
        URL to the JSON state server.
        If None, set to the default global server address.
        By default None.
    auth_client : An Auth client, optional
        An auth client with a token for the same global server, by default None
    api_version : int or 'latest', optional
        Which endpoint API version to use or 'latest'. By default, 'latest' tries to ask
        the server for which versions are available, if such functionality exists, or if not
        it defaults to the latest version for which there is a client. By default 'latest'
    ngl_url : str or None, optional
        Default neuroglancer deployment URL. Only used for V1 and later.
    """
    if server_address is None:
        server_address = default_global_server_address

    if auth_client is None:
        auth_client = AuthClient()

    auth_header = auth_client.request_header

    endpoints, api_version = _api_endpoints(api_version, server_key, server_address,
                                            jsonservice_common, jsonservice_api_versions, auth_header)

    JSONClient = client_mapping[api_version]
    return JSONClient(server_address=server_address,
                      auth_header=auth_header,
                      api_version=api_version,
                      endpoints=endpoints,
                      server_name=server_key,
                      ngl_url=ngl_url)


class JSONServiceV1(ClientBase):
    def __init__(self,
                 server_address,
                 auth_header,
                 api_version,
                 endpoints,
                 server_name,
                 ngl_url):
        super(JSONServiceV1, self).__init__(server_address,
                                            auth_header, api_version, endpoints, server_name)
        self._ngl_url = ngl_url

    @property
    def state_service_endpoint(self):
        """Endpoint URL for posting JSON state
        """
        url_mapping = self.default_url_mapping
        return self._endpoints['upload_state'].format_map(url_mapping)

    @property
    def ngl_url(self):
        return self._ngl_url

    @ngl_url.setter
    def ngl_url(self, new_ngl_url):
        self._ngl_url = new_ngl_url

    def get_state_json(self, state_id):
        """Download a Neuroglancer JSON state

        Parameters
        ----------
        state_id : int
            ID of a JSON state uploaded to the state service.

        Returns
        -------
        dict
            JSON specifying a Neuroglancer state.
        """
        url_mapping = self.default_url_mapping
        url_mapping['state_id'] = state_id
        url = self._endpoints['get_state'].format_map(url_mapping)
        response = self.session.get(url)
        handle_response(response, as_json=False)
        return json.loads(response.content)

    def upload_state_json(self, json_state, state_id=None, 
                          timestamp=None):
        """Upload a Neuroglancer JSON state

        Parameters
        ----------
        json_state : dict
            JSON-formatted Neuroglancer state
        state_id : int 
            ID of a JSON state uploaded to the state service.
            Using a state_id is an admin feature.
        timestamp: time.time
            Timestamp for json state date. Requires state_id.

        Returns
        -------
        int
            state_id of the uploaded JSON state
        """
        url_mapping = self.default_url_mapping

        if state_id is None:
            url = self._endpoints['upload_state'].format_map(url_mapping)
        else:
            url_mapping = self.default_url_mapping
            url_mapping['state_id'] = state_id
            url = self._endpoints['upload_state_w_id'].format_map(url_mapping)

        response = self.session.post(url, data=json.dumps(json_state))
        handle_response(response, as_json=False)
        response_re = re.search('.*\/(\d+)', str(response.content))
        return int(response_re.groups()[0])

    def build_neuroglancer_url(self, state_id, ngl_url=None):
        """Build a URL for a Neuroglancer deployment that will automatically retrieve specified state.
        If the datastack is specified, this is prepopulated from the info file field "viewer_site".
        If no ngl_url is specified in either the function or the client, only the JSON state url is returned.

        Parameters
        ----------
        state_id : int
            State id to retrieve
        ngl_url : str
            Base url of a neuroglancer deployment. If None, defaults to the value for the datastack or the client.
            If no value is found, only the URL to the JSON state is returned.

        Returns
        -------
        str
            The full URL requested
        """
        if ngl_url is None:
            ngl_url = self.ngl_url
        if ngl_url is None:
            ngl_url = ''
            parameter_text = ''
        elif ngl_url[-1] == '/':
            parameter_text = '?json_url='
        else:
            parameter_text = '/?json_url='

        url_mapping = self.default_url_mapping
        url_mapping['state_id'] = state_id
        get_state_url = self._endpoints['get_state'].format_map(url_mapping)
        url = ngl_url + parameter_text + get_state_url
        return url


class JSONServiceLegacy(ClientBase):
    def __init__(self,
                 server_address,
                 auth_header,
                 api_version,
                 endpoints,
                 server_name,
                 ngl_url):
        super(JSONServiceLegacy, self).__init__(server_address,
                                                auth_header, api_version, endpoints, server_name)

    @property
    def state_service_endpoint(self):
        """Endpoint URL for posting JSON state
        """
        return self._endpoints['upload_state']

    def get_state_json(self, state_id):
        """Download a Neuroglancer JSON state

        Parameters
        ----------
        state_id : int
            ID of a JSON state uploaded to the state service.

        Returns
        -------
        dict
            JSON specifying a Neuroglancer state.
        """
        url_mapping = self.default_url_mapping
        url_mapping['state_id'] = state_id
        url = self._endpoints['get_state'].format_map(url_mapping)
        response = self.session.get(url)
        handle_response(response, as_json=False)
        return json.loads(response.content)

    def upload_state_json(self, json_state):
        """Upload a Neuroglancer JSON state

        Parameters
        ----------
        json_state : dict
            JSON-formatted Neuroglancer state

        Returns
        -------
        int
            state_id of the uploaded JSON state
        """
        url_mapping = self.default_url_mapping
        url = self._endpoints['upload_state'].format_map(url_mapping)
        response = self.session.post(url, data=json.dumps(json_state))
        handle_response(response, as_json=False)
        response_re = re.search('.*\/(\d+)', str(response.content))
        return int(response_re.groups()[0])

    def build_neuroglancer_url(self, state_id, ngl_url):
        """Build a URL for a Neuroglancer deployment that will automatically retrieve specified state.

        Parameters
        ----------
        state_id : int
            State id to retrieve
        ngl_url : str
            Base url of a neuroglancer deployment. For example, 'https://neuromancer-seung-import.appspot.com'. 

        Returns
        -------
        str
            The full URL requested
        """
        url_mapping = self.default_url_mapping
        url_mapping['state_id'] = state_id
        get_state_url = self._endpoints['get_state'].format_map(url_mapping)
        if ngl_url[-1] == '/':
            spacer = '?json_url='
        else:
            spacer = '/?json_url='
        url = ngl_url + '/?json_url=' + get_state_url
        return url


client_mapping = {0: JSONServiceLegacy,
                  1: JSONServiceV1,
                  }



from .base import ClientBaseWithDataset, ClientBaseWithDatastack, ClientBase, _api_versions, _api_endpoints
from .auth import AuthClient
from .endpoints import materialization_api_versions, materialization_common
from .infoservice import InfoServiceClientV2
import requests
import time
import json
import numpy as np
from datetime import date, datetime
import pyarrow as pa
SERVER_KEY = "me_server_address"


class MEEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        if isinstance(obj, np.uint64):
            return int(obj)
        if isinstance(obj, np.int64):
            return int(obj)
        if isinstance(obj, (datetime, date)):
            return obj.isoformat()
        return json.JSONEncoder.default(self, obj)


def MaterializationClient(server_address,
                          datastack_name=None,
                          auth_client=None,
                          api_version='latest',
                          version=None,
                          verify=True):
    """ Factory for returning AnnotationClient
    Parameters
    ----------
    server_address : str 
        server_address to use to connect to (i.e. https://minniev1.microns-daf.com)
    datastack_name : str
        Name of the datastack.
    auth_client : AuthClient or None, optional
        Authentication client to use to connect to server. If None, do not use authentication.
    api_version : str or int (default: latest)
        What version of the api to use, 0: Legacy client (i.e www.dynamicannotationframework.com) 
        2: new api version, (i.e. minniev1.microns-daf.com)
        'latest': default to the most recent (current 2)
    version : default version to query
        if None will default to latest version
    Returns
    -------
    ClientBaseWithDatastack
        List of datastack names for available datastacks on the annotation engine
    """

    if auth_client is None:
        auth_client = AuthClient()

    auth_header = auth_client.request_header
    endpoints, api_version = _api_endpoints(api_version, SERVER_KEY, server_address,
                                            materialization_common, materialization_api_versions, auth_header)

    MatClient = client_mapping[api_version]
    return MatClient(server_address, auth_header, api_version,
                     endpoints, SERVER_KEY, datastack_name,
                     version=version, verify=verify)


class MaterializatonClientV2(ClientBase):
    def __init__(self, server_address, auth_header, api_version,
                 endpoints, server_name, datastack_name, version=None,
                 verify=True):
        super(MaterializatonClientV2, self).__init__(server_address,
                                                     auth_header, api_version, endpoints, server_name)

        self._datastack_name = datastack_name
        self._verify = verify
        if version is None:
            version = self.most_recent_version()
        self._version = version

    @property
    def datastack_name(self):
        return self._datastack_name

    @property
    def version(self):
        return self._version

    @version.setter
    def version(self, x):
        if int(x) in self.get_versions():
            self._version = int(x)
        else:
            raise ValueError('Version not in materialized database')

    def most_recent_version(self, datastack_name=None):
        """get the most recent version of materialization 
        for this datastack name

        Args:
            datastack_name (str, optional): datastack name to find most
            recent materialization of. 
            If None, uses the one specified in the client.
        """
        versions = self.get_versions(datastack_name=datastack_name)
        return np.max(np.array(versions))

    def get_versions(self,  datastack_name=None):
        """get versions available

        Args:
            datastack_name ([type], optional): [description]. Defaults to None.
        """
        if datastack_name is None:
            datastack_name = self.datastack_name
        endpoint_mapping = self.default_url_mapping
        endpoint_mapping["datastack_name"] = datastack_name
        url = self._endpoints["versions"].format_map(endpoint_mapping)
        response = self.session.get(url, verify=self._verify)
        self.raise_for_status(response)
        return response.json()

    def get_tables(self, datastack_name=None, version=None):
        """ Gets a list of table names for a datastack

        Parameters
        ----------
        datastack_name : str or None, optional
            Name of the datastack, by default None.
            If None, uses the one specified in the client.
            Will be set correctly if you are using the framework_client
        version: int or None, optional
            the version to query, else get the tables in the most recent version
        Returns
        -------
        list
            List of table names
        """
        if datastack_name is None:
            datastack_name = self.datastack_name
        if version is None:
            version = self.version
        endpoint_mapping = self.default_url_mapping
        endpoint_mapping["datastack_name"] = datastack_name
        endpoint_mapping["version"] = version
        # TODO fix up latest version
        url = self._endpoints["tables"].format_map(endpoint_mapping)

        response = self.session.get(url, verify=self._verify)
        self.raise_for_status(response)
        return response.json()

    def get_annotation_count(self, table_name: str,
                             datastack_name=None,
                             version=None):
        """ Get number of annotations in a table

        Parameters
        ----------
        table_name (str): 
            name of table to mark for deletion
        datastack_name: str or None, optional,
            Name of the datastack_name. If None, uses the one specified in the client.
        version: int or None, optional
            the version to query, else get the tables in the most recent version
        Returns
        -------
        int
            number of annotations
        """
        if datastack_name is None:
            datastack_name = self.datastack_name

        endpoint_mapping = self.default_url_mapping
        endpoint_mapping["datastack_name"] = datastack_name
        endpoint_mapping["table_name"] = table_name

        url = self._endpoints["table_count"].format_map(endpoint_mapping)

        response = self.session.get(url, verify=self._verify)
        self.raise_for_status(response)
        return response.json()

    def get_version_metadata(self, version: int = None, datastack_name: str = None):
        """get metadata about a version

        Args:
            version (int, optional): version number to get metadata about. Defaults to client default version.
            datastack_name (str, optional): datastack to query. Defaults to client default datastack.
        """
        if datastack_name is None:
            datastack_name = self.datastack_name
        if version is None:
            version = self.version

        endpoint_mapping = self.default_url_mapping
        endpoint_mapping["datastack_name"] = datastack_name
        endpoint_mapping["version"] = version
        url = self._endpoints["version_metadata"].format_map(endpoint_mapping)
        response = self.session.get(url, verify=self._verify)
        self.raise_for_status(response)
        return response.json()

    def get_timestamp(self, version: int = None, datastack_name: str = None):
        """Get datetime.datetime timestamp for a materialization version.

        Parameters
        ----------
        version : int or None, optional
            Materialization version, by default None. If None, defaults to the value set in the client.
        datastack_name : str or None, optional
            Datastack name, by default None. If None, defaults to the value set in the client.

        Returns
        -------
        datetime.datetime
            Datetime when the materialization version was frozen.
        """
        meta = self.get_version_metadata(
            version=version, datastack_name=datastack_name)
        return datetime.strptime(meta['time_stamp'], '%Y-%m-%dT%H:%M:%S.%f')

    def get_table_metadata(self, table_name: str, datastack_name=None):
        """ Get metadata about a table

        Parameters
        ----------
        table_name (str): 
            name of table to mark for deletion
        datastack_name: str or None, optional,
            Name of the datastack_name.
            If None, uses the one specified in the client.


        Returns
        -------
        json
            metadata about table
        """
        if datastack_name is None:
            datastack_name = self.datastack_name

        endpoint_mapping = self.default_url_mapping
        endpoint_mapping["datastack_name"] = datastack_name
        endpoint_mapping["table_name"] = table_name

        url = self._endpoints["metadata"].format_map(endpoint_mapping)

        response = self.session.get(url, verify=self._verify)
        self.raise_for_status(response)
        return response.json()

    # def get_annotation(self, table_name, annotation_ids,
    #                    materialization_version=None,
    #                    datastack_name=None):
    #     """ Retrieve an annotation or annotations by id(s) and table name.

    #     Parameters
    #     ----------
    #     table_name : str
    #         Name of the table
    #     annotation_ids : int or iterable
    #         ID or IDS of the annotation to retreive
    #     materialization_version: int or None
    #         materialization version to use
    #         If None, uses the one specified in the client
    #     datastack_name : str or None, optional
    #         Name of the datastack_name.
    #         If None, uses the one specified in the client.
    #     Returns
    #     -------
    #     list
    #         Annotation data
    #     """
    #     if materialization_version is None:
    #         materialization_version = self.version
    #     if datastack_name is None:
    #         datastack_name = self.datastack_name

    #     endpoint_mapping = self.default_url_mapping
    #     endpoint_mapping["datastack_name"] = datastack_name
    #     endpoint_mapping["table_name"] = table_name
    #     endpoint_mapping["version"] = materialization_version
    #     url = self._endpoints["annotations"].format_map(endpoint_mapping)
    #     try:
    #         iter(annotation_ids)
    #     except TypeError:
    #         annotation_ids = [annotation_ids]

    #     params = {
    #         'annotation_ids': ",".join([str(a) for a in annotation_ids])
    #     }
    #     response = self.session.get(url, params=params)
    #     self.raise_for_status(response)
    #     return response.json()

    def query_table(self,
                    table: str,
                    filter_in_dict=None,
                    filter_out_dict=None,
                    filter_equal_dict=None,
                    filter_spatial=None,
                    join_args=None,
                    select_columns=None,
                    offset: int = None,
                    limit: int = None,
                    datastack_name: str = None,
                    materialization_version: int = None):
        """generic query on materialization tables

        Args:
            table: 'str'

            filter_in_dict (dict , optional): 
                keys are column names, values are allowed entries.
                Defaults to None.
            filter_out_dict (dict, optional): 
                keys are column names, values are not allowed entries.
                Defaults to None.
            filter_equal_dict (dict, optional): 
                inner layer: keys are column names, values are specified entry.
                Defaults to None.
            offset (int, optional): offset in query result
            limit (int, optional): maximum results to return (server will set upper limit, see get_server_config)
            select_columns (list of str, optional): columns to select. Defaults to None.
            suffixes: (list[str], optional): suffixes to use on duplicate columns
            offset (int, optional): result offset to use. Defaults to None.
                will only return top K results. 
            datastack_name (str, optional): datastack to query. 
                If None defaults to one specified in client. 
            materialization_version (int, optional): version to query. 
                If None defaults to one specified in client.
        Returns:
        pd.DataFrame: a pandas dataframe of results of query

        """
        if materialization_version is None:
            materialization_version = self.version
        if datastack_name is None:
            datastack_name = self.datastack_name

        endpoint_mapping = self.default_url_mapping
        endpoint_mapping["datastack_name"] = datastack_name
        endpoint_mapping["version"] = materialization_version
        data = {}
        query_args = {}
        if type(table) == str:
            tables = [table]
        if len(tables) == 1:
            assert(type(tables[0]) == str)
            endpoint_mapping["table_name"] = tables[0]
            single_table = True
            url = self._endpoints["simple_query"].format_map(endpoint_mapping)
        else:
            single_table = False
            data['tables'] = tables
            url = self._endpoints["join_query"].format_map(endpoint_mapping)

        if filter_in_dict is not None:
            data['filter_in_dict'] = {table: filter_in_dict}
        if filter_out_dict is not None:
            data['filter_notin_dict'] = {table: filter_out_dict}
        if filter_equal_dict is not None:
            data['filter_equal_dict'] = {table: filter_equal_dict}
        if select_columns is not None:
            data['select_columns'] = select_columns
        if offset is not None:
            data['offset'] = offset
        if limit is not None:
            assert(limit > 0)
            data['limit'] = limit
        response = self.session.post(url, data=json.dumps(data, cls=MEEncoder),
                                     headers={
                                         'Content-Type': 'application/json'},
                                     verify=self._verify)
        self.raise_for_status(response)
        return pa.deserialize(response.content)

    def join_query(self,
                   tables,
                   filter_in_dict=None,
                   filter_out_dict=None,
                   filter_equal_dict=None,
                   filter_spatial=None,
                   join_args=None,
                   select_columns=None,
                   offset: int = None,
                   limit: int = None,
                   suffixes: list = None,
                   datastack_name: str = None,
                   materialization_version: int = None):
        """generic query on materialization tables

        Args:
            tables: list of lists with length 2 or 'str'
                list of two lists: first entries are table names, second
                                   entries are the columns used for the join
            filter_in_dict (dict of dicts, optional): 
                outer layer: keys are table names
                inner layer: keys are column names, values are allowed entries.
                Defaults to None.
            filter_out_dict (dict of dicts, optional): 
                outer layer: keys are table names
                inner layer: keys are column names, values are not allowed entries.
                Defaults to None.
            filter_equal_dict (dict of dicts, optional): 
                outer layer: keys are table names
                inner layer: keys are column names, values are specified entry.
                Defaults to None.
            select_columns (list of str, optional): columns to select. Defaults to None.
            offset (int, optional): result offset to use. Defaults to None.
                will only return top K results. 
            limit (int, optional): maximum results to return (server will set upper limit, see get_server_config)
            suffixes (list[str], optional): suffixes to use for duplicate columns same order as tables 
            datastack_name (str, optional): datastack to query. 
                If None defaults to one specified in client. 
            materialization_version (int, optional): version to query. 
                If None defaults to one specified in client.
        Returns:
        pd.DataFrame: a pandas dataframe of results of query

        """
        if materialization_version is None:
            materialization_version = self.version
        if datastack_name is None:
            datastack_name = self.datastack_name

        endpoint_mapping = self.default_url_mapping
        endpoint_mapping["datastack_name"] = datastack_name
        endpoint_mapping["version"] = materialization_version
        data = {}
        query_args = {}

        data['tables'] = tables
        url = self._endpoints["join_query"].format_map(endpoint_mapping)

        if filter_in_dict is not None:
            data['filter_in_dict'] = filter_in_dict
        if filter_out_dict is not None:
            data['filter_notin_dict'] = filter_out_dict
        if filter_equal_dict is not None:
            data['filter_equal_dict'] = filter_equal_dict
        if select_columns is not None:
            data['select_columns'] = select_columns
        if offset is not None:
            data['offset'] = offset
        if suffixes is not None:
            data['suffixes'] = suffixes
        if limit is not None:
            assert(limit > 0)
            data['limit'] = limit
        response = self.session.post(url, data=json.dumps(data, cls=MEEncoder),
                                     headers={
                                         'Content-Type': 'application/json'},
                                     verify=self._verify)
        self.raise_for_status(response)
        return pa.deserialize(response.content)


client_mapping = {2: MaterializatonClientV2,
                  'latest': MaterializatonClientV2}



from urllib.parse import urlparse


def format_precomputed_neuroglancer(objurl):
    qry = urlparse(objurl)
    if qry.scheme == 'gs':
        objurl_out = f'precomputed://{objurl}'
    elif qry.scheme == 'http' or qry.scheme == 'https':
        objurl_out = f'precomputed://gs://{qry.path[1:]}'
    else:
        objurl_out = None
    return objurl_out


def format_precomputed_https(objurl):
    qry = urlparse(objurl)
    if qry.scheme == 'gs':
        objurl_out = f'precomputed://https://storage.googleapis.com/{qry.path[1:]}'
    elif qry.scheme == 'http' or qry.scheme == 'https':
        objurl_out = f'precomputed://{objurl}'
    else:
        objurl_out = None
    return objurl_out


def format_graphene(objurl):
    qry = urlparse(objurl)
    if qry.scheme == 'http' or qry.scheme == 'https':
        objurl_out = f'graphene://{objurl}'
    elif qry.scheme == 'graphene':
        objurl_out = objurl
    else:
        objurl_out = None
    return objurl_out


def format_cloudvolume(objurl):
    qry = urlparse(objurl)
    if qry.scheme == 'graphene':
        return format_graphene(objurl)
    elif qry.scheme == 'gs' or qry.scheme == 'http' or qry.scheme == 'https':
        return format_precomputed_https(objurl)
    else:
        return None


def format_raw(objurl):
    return objurl


# No reformatting
output_map_raw = {}

# Use precomputed://gs:// links for neuroglancer, but use precomputed://https://storage.googleapis.com links in cloudvolume
output_map_precomputed = {'raw': format_raw,
                          'cloudvolume': format_precomputed_https,
                          'neuroglancer': format_precomputed_neuroglancer}

# Use graphene://https:// links for both neuroglancer and cloudvolume
output_map_graphene = {'raw': format_raw,
                       'cloudvolume': format_graphene,
                       'neuroglancer': format_graphene}



from .base import ClientBaseWithDataset, ClientBaseWithDatastack, _api_versions, _api_endpoints, handle_response
from .auth import AuthClient
from .endpoints import infoservice_common, infoservice_api_versions, default_global_server_address
from .format_utils import output_map_raw, output_map_precomputed, output_map_graphene, format_raw
import requests
from warnings import warn

SERVER_KEY = "i_server_address"


def InfoServiceClient(server_address=None,
                      datastack_name=None,
                      auth_client=None,
                      api_version='latest',
                      ):
    if server_address is None:
        server_address = default_global_server_address

    if auth_client is None:
        auth_client = AuthClient()

    auth_header = auth_client.request_header
    endpoints, api_version = _api_endpoints(api_version, SERVER_KEY, server_address,
                                            infoservice_common, infoservice_api_versions, auth_header)

    InfoClient = client_mapping[api_version]
    return InfoClient(server_address,
                      auth_header,
                      api_version,
                      endpoints,
                      SERVER_KEY,
                      datastack_name,
                      )


class InfoServiceClientLegacy(ClientBaseWithDataset):
    def __init__(self,
                 server_address,
                 auth_header,
                 api_version,
                 endpoints,
                 server_name,
                 dataset_name):
        super(InfoServiceClientLegacy, self).__init__(server_address,
                                                      auth_header,
                                                      api_version,
                                                      endpoints,
                                                      server_name,
                                                      dataset_name)
        self.info_cache = dict()

    def get_datasets(self):
        """Query which datasets are available at the info service

        Returns
        -------
        list
            List of dataset names
        """
        endpoint_mapping = self.default_url_mapping
        url = self._endpoints['datasets'].format_map(endpoint_mapping)

        response = self.session.get(url)
        return handle_response(response)

    def get_dataset_info(self, dataset_name=None, use_stored=True):
        """Gets the info record for a dataset

        Parameters
        ----------
        dataset_name : str, optional
            Dataset to look up. If None, uses the one specified by the client. By default None
        use_stored : bool, optional
            If True and the information has already been queried for that dataset, then uses the cached version. If False, re-queries the infromation. By default True

        Returns
        -------
        dict or None
            The complete info record for the dataset
        """
        if dataset_name is None:
            dataset_name = self.dataset_name
        if dataset_name is None:
            raise ValueError('No Dataset set')

        if (not use_stored) or (dataset_name not in self.info_cache):
            endpoint_mapping = self.default_url_mapping
            endpoint_mapping['dataset_name'] = dataset_name
            url = self._endpoints['dataset_info'].format_map(endpoint_mapping)

            response = self.session.get(url)
            self.info_cache[dataset_name] = handle_response(response)

        return self.info_cache.get(dataset_name, None)

    def _get_property(self, info_property, dataset_name=None, use_stored=True, format_for='raw', output_map=output_map_raw):
        if dataset_name is None:
            dataset_name = self.dataset_name
        if dataset_name is None:
            raise ValueError('No Dataset set')

        self.get_dataset_info(dataset_name=dataset_name, use_stored=use_stored)
        return output_map.get(format_for, format_raw)(self.info_cache[dataset_name].get(info_property, None))

    def annotation_endpoint(self, dataset_name=None, use_stored=True):
        """AnnotationEngine endpoint for a dataset.

        Parameters
        ----------
        dataset_name : str or None, optional
            Name of the dataset to look up. If None, uses the value specified by the client. Default is None.
        use_stored : bool, optional
            If True, uses the cached value if available. If False, re-queries the InfoService. Default is True.

        Returns
        -------
        str
            Location of the AnnotationEngine
        """
        return self._get_property('annotation_engine_endpoint',
                                  dataset_name=dataset_name,
                                  use_stored=use_stored,
                                  output_map=output_map_raw)

    def flat_segmentation_source(self, dataset_name=None, use_stored=True, format_for='raw'):
        """Cloud path to the flat segmentation for the dataset

        Parameters
        ----------
        dataset_name : str or None, optional
            Name of the dataset to look up. If None, uses the value specified by the client. Default is None.
        use_stored : bool, optional
            If True, uses the cached value if available. If False, re-queries the InfoService. Default is True.
        format_for : 'raw', 'cloudvolume', or 'neuroglancer', optional
            Formats the path for different uses.
            If 'raw' (default), the path in the InfoService is passed along.
            If 'cloudvolume', a "precomputed://gs://" type path is converted to a full https URL.
            If 'neuroglancer', a full https URL is converted to a "precomputed://gs://" type path.

        Returns
        -------
        str
            Formatted cloud path to the flat segmentation
        """
        return self._get_property('flat_segmentation_source',
                                  dataset_name=dataset_name,
                                  use_stored=use_stored,
                                  format_for=format_for,
                                  output_map=output_map_precomputed)

    def image_source(self, dataset_name=None, use_stored=True, format_for='raw'):
        """Cloud path to the imagery for the dataset

        Parameters
        ----------
        dataset_name : str or None, optional
            Name of the dataset to look up. If None, uses the value specified by the client. Default is None.
        use_stored : bool, optional
            If True, uses the cached value if available. If False, re-queries the InfoService. Default is True.
        format_for : 'raw', 'cloudvolume', or 'neuroglancer', optional
            Formats the path for different uses.
            If 'raw' (default), the path in the InfoService is passed along.
            If 'cloudvolume', a "precomputed://gs://" type path is converted to a full https URL.
            If 'neuroglancer', a full https URL is converted to a "precomputed://gs://" type path.

        Returns
        -------
        str
            Formatted cloud path to the flat segmentation
        """
        return self._get_property('image_source',
                                  dataset_name=dataset_name,
                                  use_stored=use_stored,
                                  format_for=format_for,
                                  output_map=output_map_precomputed)

    def synapse_segmentation_source(self, dataset_name=None,
                                    use_stored=True, format_for='raw'):
        """Cloud path to the synapse segmentation for a dataset

        Parameters
        ----------
        dataset_name : str or None, optional
            Name of the dataset to look up. If None, uses the value specified by the client. Default is None.
        use_stored : bool, optional
            If True, uses the cached value if available. If False, re-queries the InfoService. Default is True.
        format_for : 'raw', 'cloudvolume', or 'neuroglancer', optional
            Formats the path for different uses.
            If 'raw' (default), the path in the InfoService is passed along.
            If 'cloudvolume', a "precomputed://gs://" type path is converted to a full https URL.
            If 'neuroglancer', a full https URL is converted to a "precomputed://gs://" type path.

        Returns
        -------
        str
            Formatted cloud path to the synapse segmentation
        """
        return self._get_property('synapse_segmentation_source',
                                  dataset_name=dataset_name,
                                  use_stored=use_stored,
                                  format_for=format_for,
                                  output_map=output_map_precomputed)

    def supervoxel_source(self, dataset_name=None, use_stored=True, format_for='raw'):
        """Cloud path to the supervoxel segmentation for a dataset

        Parameters
        ----------
        dataset_name : str or None, optional
            Name of the dataset to look up. If None, uses the value specified by the client. Default is None.
        use_stored : bool, optional
            If True, uses the cached value if available. If False, re-queries the InfoService. Default is True.
        format_for : 'raw', 'cloudvolume', or 'neuroglancer', optional
            Formats the path for different uses.
            If 'raw' (default), the path in the InfoService is passed along.
            If 'cloudvolume', a "precomputed://gs://" type path is converted to a full https URL.
            If 'neuroglancer', a full https URL is converted to a "precomputed://gs://" type path.

        Returns
        -------
        str
            Formatted cloud path to the supervoxel segmentation
        """
        return self._get_property('pychunkedgraph_supervoxel_source',
                                  dataset_name=dataset_name, use_stored=use_stored,
                                  format_for=format_for, output_map=output_map_precomputed)

    def pychunkedgraph_endpoint(self, dataset_name=None, use_stored=True):
        return self._get_property('pychunkgraph_endpoint',
                                  dataset_name=dataset_name,
                                  use_stored=use_stored,
                                  output_map=output_map_raw)

    def pychunkgraph_endpoint(self, **kwargs):
        warn('Please use ''pychunkedgraph_endpoint''', DeprecationWarning)
        return self.pychunkedgraph_endpoint(**kwargs)

    def pychunkedgraph_segmentation_source(self, dataset_name=None,
                                           use_stored=True, format_for='raw'):
        return self._get_property('pychunkgraph_segmentation_source',
                                  dataset_name=dataset_name,
                                  use_stored=use_stored,
                                  format_for=format_for,
                                  output_map=output_map_graphene)

    def pychunkgraph_segmentation_source(self, **kwargs):
        warn('Please use ''pychunkedgraph_segmentation_source'' in the future.',
             DeprecationWarning)
        return self.pychunkedgraph_segmentation_source(**kwargs)

    def pychunkedgraph_viewer_source(self, **kwargs):
        warn('Use ''graphene_source'' instead', DeprecationWarning)
        return self.graphene_source(**kwargs)

    def graphene_source(self, dataset_name=None,
                        use_stored=True, format_for='raw'):
        """Cloud path to the chunkgraph-backed Graphene segmentation for a dataset

        Parameters
        ----------
        dataset_name : str or None, optional
            Name of the dataset to look up. If None, uses the value specified by the client. Default is None.
        use_stored : bool, optional
            If True, uses the cached value if available. If False, re-queries the InfoService. Default is True.
        format_for : 'raw', 'cloudvolume', or 'neuroglancer', optional
            Formats the path for different uses.
            If 'raw' (default), the path in the InfoService is passed along.
            If 'cloudvolume', a "graphene://https://" type path is used
            If 'neuroglancer', a "graphene://https://" type path is used, as needed by Neuroglancer.

        Returns
        -------
        str
            Formatted cloud path to the Graphene segmentation
        """
        return self._get_property('graphene_source',
                                  dataset_name=dataset_name,
                                  use_stored=use_stored,
                                  format_for=format_for,
                                  output_map=output_map_graphene)

    def refresh_stored_data(self):
        """Reload the stored info values from the server.
        """
        for ds in self.info_cache.keys():
            self.get_dataset_info(dataset_name=ds, use_stored=False)


class InfoServiceClientV2(ClientBaseWithDatastack):
    def __init__(self,
                 server_address,
                 auth_header,
                 api_version,
                 endpoints,
                 server_name,
                 datastack_name):
        super(InfoServiceClientV2, self).__init__(server_address,
                                                  auth_header,
                                                  api_version,
                                                  endpoints,
                                                  server_name,
                                                  datastack_name)
        self.info_cache = dict()
        if datastack_name is not None:
            ds_info = self.get_datastack_info(datastack_name=datastack_name)
            self._aligned_volume_name = ds_info['aligned_volume']['id']
            self._aligned_volume_id = ds_info['aligned_volume']['name']
        else:
            self._aligned_volume_name = None
            self._aligned_volume_id = None

    @property
    def aligned_volume_name(self):
        return self._aligned_volume_name

    @property
    def aligned_volume_id(self):
        return self._aligned_volume_id

    def get_datastacks(self):
        """Query which datastacks are available at the info service

        Returns
        -------
        list
            List of datastack names
        """
        endpoint_mapping = self.default_url_mapping
        url = self._endpoints['datastacks'].format_map(endpoint_mapping)

        response = self.session.get(url)
        return handle_response(response)

    def get_datastack_info(self, datastack_name=None, use_stored=True):
        """Gets the info record for a datastack

        Parameters
        ----------
        datastack_name : str, optional
            datastack to look up. If None, uses the one specified by the client. By default None
        use_stored : bool, optional
            If True and the information has already been queried for that datastack, then uses the cached version. If False, re-queries the infromation. By default True

        Returns
        -------
        dict or None
            The complete info record for the datastack
        """
        if datastack_name is None:
            datastack_name = self.datastack_name
        if datastack_name is None:
            raise ValueError('No Dataset set')

        if (not use_stored) or (datastack_name not in self.info_cache):
            endpoint_mapping = self.default_url_mapping
            endpoint_mapping['datastack_name'] = datastack_name
            url = self._endpoints['datastack_info'].format_map(
                endpoint_mapping)

            response = self.session.get(url)
            self.raise_for_status(response)

            self.info_cache[datastack_name] = handle_response(response)

        return self.info_cache.get(datastack_name, None)

    def _get_property(self, info_property, datastack_name=None, use_stored=True, format_for='raw', output_map=output_map_raw):
        if datastack_name is None:
            datastack_name = self.datastack_name
        if datastack_name is None:
            raise ValueError('No Dataset set')

        self.get_datastack_info(
            datastack_name=datastack_name, use_stored=use_stored)
        value = self.info_cache[datastack_name].get(info_property, None)
        return output_map.get(format_for, format_raw)(value)

    def get_aligned_volumes(self):
        endpoint_mapping = self.default_url_mapping
        url = self._endpoints['aligned_volumes'].format_map(endpoint_mapping)
        response = self.session.get(url)
        return handle_response(response)

    def get_aligned_volume_info(self, datastack_name: str = None, use_stored=True):
        """Gets the info record for a aligned_volume

        Parameters
        ----------
        datastack_name : str, optional
            datastack_name to look up. If None, uses the one specified by the client. By default None
        use_stored : bool, optional
            If True and the information has already been queried for that dataset, then uses the cached version. If False, re-queries the infromation. By default True

        Returns
        -------
        dict or None
            The complete info record for the aligned_volume
        """
        return self._get_property('aligned_volume',
                                  datastack_name=datastack_name,
                                  use_stored=use_stored)

    def get_aligned_volume_info_by_id(self, aligned_volume_id: int = None, use_stored=True):
        if aligned_volume_id is None:
            aligned_volume_id = self._aligned_volume_id
        if aligned_volume_id is None:
            raise ValueError(
                "Must specify aligned_volume_id or provide datastack_name in init")

        endpoint_mapping = self.default_url_mapping
        endpoint_mapping['aligned_volume_id'] = aligned_volume_id
        url = self._endpoints['aligned_volume_by_id'].format_map(
            endpoint_mapping)

        response = self.session.get(url)
        return handle_response(response)

    def local_server(self, datastack_name=None, use_stored=True):
        return self._get_property('local_server',
                                  datastack_name=datastack_name,
                                  use_stored=use_stored,
                                  output_map=output_map_raw)

    def annotation_endpoint(self, datastack_name=None, use_stored=True):
        """AnnotationEngine endpoint for a dataset.

        Parameters
        ----------
        datastack_name : str or None, optional
            Name of the datastack to look up. If None, uses the value specified by the client. Default is None.
        use_stored : bool, optional
            If True, uses the cached value if available. If False, re-queries the InfoService. Default is True.

        Returns
        -------
        str
            Location of the AnnotationEngine
        """
        local_server = self.local_server(
            datastack_name=datastack_name, use_stored=use_stored)

        return local_server + "/annotation"

    def image_source(self, datastack_name=None, use_stored=True, format_for='raw'):
        """Cloud path to the imagery for the dataset

        Parameters
        ----------
        datastack_name : str or None, optional
            Name of the datastack to look up. If None, uses the value specified by the client. Default is None.
        use_stored : bool, optional
            If True, uses the cached value if available. If False, re-queries the InfoService. Default is True.
        format_for : 'raw', 'cloudvolume', or 'neuroglancer', optional
            Formats the path for different uses.
            If 'raw' (default), the path in the InfoService is passed along.
            If 'cloudvolume', a "precomputed://gs://" type path is converted to a full https URL.
            If 'neuroglancer', a full https URL is converted to a "precomputed://gs://" type path.

        Returns
        -------
        str
            Formatted cloud path to the flat segmentation
        """

        av_info = self.get_aligned_volume_info(datastack_name=datastack_name,
                                               use_stored=use_stored)
        return av_info['image_source']

    def synapse_segmentation_source(self, datastack_name=None,
                                    use_stored=True, format_for='raw'):
        """Cloud path to the synapse segmentation for a dataset

        Parameters
        ----------
        datastack_name : str or None, optional
            Name of the dataset to look up. If None, uses the value specified by the client. Default is None.
        use_stored : bool, optional
            If True, uses the cached value if available. If False, re-queries the InfoService. Default is True.
        format_for : 'raw', 'cloudvolume', or 'neuroglancer', optional
            Formats the path for different uses.
            If 'raw' (default), the path in the InfoService is passed along.
            If 'cloudvolume', a "precomputed://gs://" type path is converted to a full https URL.
            If 'neuroglancer', a full https URL is converted to a "precomputed://gs://" type path.

        Returns
        -------
        str
            Formatted cloud path to the synapse segmentation
        """
        return self._get_property('synapse_segmentation_source',
                                  datastack_name=datastack_name,
                                  use_stored=use_stored,
                                  format_for=format_for,
                                  output_map=output_map_precomputed)

    def segmentation_source(self, datastack_name=None, format_for='raw', use_stored=True):
        """Cloud path to the chunkgraph-backed Graphene segmentation for a dataset

        Parameters
        ----------
        datastack_name : str or None, optional
            Name of the datastack to look up. If None, uses the value specified by the client. Default is None.
        use_stored : bool, optional
            If True, uses the cached value if available. If False, re-queries the InfoService. Default is True.
        format_for : 'raw', 'cloudvolume', or 'neuroglancer', optional
            Formats the path for different uses.
            If 'raw' (default), the path in the InfoService is passed along.
            If 'cloudvolume', a "graphene://https://" type path is used
            If 'neuroglancer', a "graphene://https://" type path is used, as needed by Neuroglancer.

        Returns
        -------
        str
            Formatted cloud path to the Graphene segmentation
        """
        return self._get_property('segmentation_source',
                                  datastack_name=datastack_name,
                                  use_stored=use_stored,
                                  output_map=output_map_raw)

    def refresh_stored_data(self):
        """Reload the stored info values from the server.
        """
        for ds in self.info_cache.keys():
            self.get_datastack_info(datastack_name=ds, use_stored=False)

    def viewer_site(self, datastack_name=None, use_stored=True):
        """Get the base Neuroglancer URL for the dataset
        """
        return self._get_property('viewer_site',
                                  datastack_name=datastack_name,
                                  use_stored=use_stored,
                                  format_for='raw')


client_mapping = {0: InfoServiceClientLegacy,
                  2: InfoServiceClientV2,
                  'latest': InfoServiceClientV2}




default_global_server_address = "https://globalv1.daf-apis.com"

# -------------------------------
# ------ AnnotationEngine endpoints
# -------------------------------

annotation_common = {}

anno_legacy = "{ae_server_address}/annotation"
annotation_endpoints_legacy = {
    "datasets": anno_legacy + "/datasets",
    "table_names": anno_legacy + "/dataset/{dataset_name}",
    "existing_annotation": anno_legacy + "/dataset/{dataset_name}/{table_name}/{annotation_id}",
    "new_annotation": anno_legacy + "/dataset/{dataset_name}/{table_name}",
}

anno_v2 = "{ae_server_address}/annotation/api/v2"
annotation_endpoints_v2 = {
    "tables": anno_v2 + "/aligned_volume/{aligned_volume_name}/table",
    "table_info": anno_v2 + "/aligned_volume/{aligned_volume_name}/table/{table_name}",
    "annotations": anno_v2 + "/aligned_volume/{aligned_volume_name}/table/{table_name}/annotations",
    "table_count": anno_v2 + "/aligned_volume/{aligned_volume_name}/table/{table_name}/count",
}


materialization_common = {}
mat_v2_api = "{me_server_address}/materialize/api/v2"
materialization_endpoints_v2 = {
    "simple_query": mat_v2_api + "/datastack/{datastack_name}/version/{version}/table/{table_name}/query",
    "join_query": mat_v2_api + "/datastack/{datastack_name}/version/{version}/query",
    "annotations": mat_v2_api + "/datastack/{datastack_name}/version/{version}/table/{table_name}",
    "table_count": mat_v2_api + "/datastack/{datastack_name}/version/{version}/table/{table_name}/count",
    "versions": mat_v2_api + "/datastack/{datastack_name}/versions",
    "version_metadata": mat_v2_api + "/datastack/{datastack_name}/version/{version}",
    "tables": mat_v2_api + "/datastack/{datastack_name}/version/{version}/tables",
    "metadata": mat_v2_api + "/datastack/{datastack_name}/version/{version}/table/{table_name}/metadata",
}
materialization_api_versions = {2: materialization_endpoints_v2}
annotation_api_versions = {0: annotation_endpoints_legacy,
                           2: annotation_endpoints_v2}

# -------------------------------
# ------ Infoservice endpoints
# -------------------------------

infoservice_common = {}

info_v1 = "{i_server_address}/info/api"
infoservice_endpoints_v1 = {
    "datasets": info_v1 + "/datasets",
    "dataset_info": info_v1 + "/dataset/{dataset_name}",
}

info_v2 = "{i_server_address}/info/api/v2"
infoservice_endpoints_v2 = {
    "aligned_volumes": info_v2 + "/aligned_volume",
    "aligned_volume_info": info_v2 + "/aligned_volume/{aligned_volume_name}",
    "aligned_volume_by_id": info_v2 + "/aligned_volume/id/{aligned_volume_id}",
    "datastacks": info_v2 + "/datastacks",
    "datastack_info": info_v2 + "/datastack/full/{datastack_name}",
}

infoservice_api_versions = {1: infoservice_endpoints_v1,
                            2: infoservice_endpoints_v2}

# -------------------------------
# ------ Pychunkedgraph endpoints
# -------------------------------

pcg_common = "{cg_server_address}/segmentation"
chunkedgraph_endpoints_common = {
    "get_api_versions": pcg_common + "/api/versions",
    'info': pcg_common + "/table/{table_id}/info",
}

pcg_legacy = "{cg_server_address}/segmentation/1.0"
chunkedgraph_endpoints_legacy = {
    # "handle_table": "{cg_server_address}/segmentation/1.0/table",
    "handle_root": pcg_legacy + "/{table_id}/graph/root",
    "handle_children": pcg_legacy + "/segment/{node_id}/children",
    # "info": pcg_legacy + "/{table_id}/info",
    "leaves_from_root": pcg_legacy + "/{table_id}/segment/{root_id}/leaves",
    "merge_log": pcg_legacy + "/{table_id}/segment/{root_id}/merge_log",
    "change_log":  pcg_legacy + "/{table_id}/segment/{root_id}/change_log",
    "contact_sites": pcg_legacy + "/{table_id}/segment/{root_id}/contact_sites",
    "cloudvolume_path": "graphene://" + pcg_legacy + "/{table_id}",
}

pcg_v1 = "{cg_server_address}/segmentation/api/v1"
pcg_meshing_v1 = "{cg_server_address}/meshing/api/v1"
chunkedgraph_endpoints_v1 = {
    'handle_root': pcg_v1 + "/table/{table_id}/node/{supervoxel_id}/root",
    "handle_roots": pcg_v1 + "/table/{table_id}/roots",
    'handle_children': pcg_v1 + "/table/{table_id}/node/{root_id}/childen",
    'leaves_from_root': pcg_v1 + "/table/{table_id}/node/{root_id}/leaves",
    'do_merge': pcg_v1 + "/table/{table_id}/merge",
    'get_roots': pcg_v1 + "/table/{table_id}/roots_binary",
    'merge_log': pcg_v1 + "/table/{table_id}/root/{root_id}/merge_log",
    'change_log': pcg_v1 + "/table/{table_id}/root/{root_id}/change_log",
    'contact_sites': pcg_v1 + "/table/{table_id}/node/{root_id}/contact_sites",
    'contact_sites_pairwise': pcg_v1 + "/table/{table_id}/contact_sites_pair/{root_id_1}/{root_id_2}",
    'cloudvolume_path': "graphene://" + pcg_v1 + "/{table_id}",
    'find_path': pcg_v1 + "/table/{table_id}/graph/find_path",
    'lvl2_graph': pcg_v1 + "/table/{table_id}/node/{root_id}/lvl2_graph",
    'remesh_level2_chunks': pcg_meshing_v1 + "/table/{table_id}/remeshing",
    'get_subgraph': pcg_v1 + "/table/{table_id}/node/{root_id}/subgraph"
}

chunkedgraph_api_versions = {0: chunkedgraph_endpoints_legacy,
                             1: chunkedgraph_endpoints_v1,
                             }

# -------------------------------
# ------ EMAnnotationSchemas endpoints
# -------------------------------
schema_common = "{emas_server_address}/schema"
schema_endpoints_common = {
     "get_api_versions": schema_common + "/versions",
}

schema_v1 = "{emas_server_address}/schema"
schema_endpoints_v1 = {
    "schema": schema_v1 + "/type",
    "schema_definition": schema_v1 + "/type/{schema_type}"
}
schema_v2 = "{emas_server_address}/schema/api/v2"
schema_endpoints_v2 = {
    "schema": schema_v2 + "/type",
    "schema_definition": schema_v2 + "/type/{schema_type}"
}

schema_api_versions = {1: schema_endpoints_v1,
                       2: schema_endpoints_v2}

# -------------------------------
# ------ StateServer endpoints
# -------------------------------

jsonservice_common = {}

json_v1 = "{json_server_address}/nglstate/api/v1"
jsonservice_endpoints_v1 = {
    "upload_state": json_v1 + "/post",
    "upload_state_w_id": json_v1 + "/post/{state_id}",
    "get_state": json_v1 + "/{state_id}",
    'get_state_raw': json_v1 + "/raw/{state_id}",
}

json_legacy = "{json_server_address}/nglstate"
jsonservice_endpoints_legacy = {
    "upload_state": json_legacy + "/post",
    "get_state": json_legacy + "/{state_id}",
    'get_state_raw': json_legacy + "/raw/{state_id}",
}

jsonservice_api_versions = {0: jsonservice_endpoints_legacy,
                            1: jsonservice_endpoints_v1,
                            }

# -------------------------------
# ------ Auth endpoints
# -------------------------------

auth_common = {}

v1_auth = "{auth_server_address}/auth/api/v1"
auth_endpoints_v1 = {
    "refresh_token": v1_auth + "/refresh_token"
}

auth_api_versions = {1: auth_endpoints_v1,
                     }



from .base import ClientBaseWithDataset, ClientBaseWithDatastack, ClientBase, _api_versions, _api_endpoints, handle_response
from .auth import AuthClient
from .endpoints import annotation_common, annotation_api_versions
from .infoservice import InfoServiceClientV2
import requests
import time
import json
import numpy as np
from datetime import date, datetime

SERVER_KEY = "ae_server_address"


class AEEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        if isinstance(obj, np.uint64):
            return int(obj)
        if isinstance(obj, (datetime, date)):
            return obj.isoformat()
        return json.JSONEncoder.default(self, obj)


def AnnotationClient(server_address,
                     dataset_name=None,
                     aligned_volume_name=None,
                     auth_client=None,
                     api_version='latest'):
    """ Factory for returning AnnotationClient
    Parameters
    ----------
    server_address : str 
        server_address to use to connect to (i.e. https://minniev1.microns-daf.com)
    datastack_name : str
        Name of the datastack.
    auth_client : AuthClient or None, optional
        Authentication client to use to connect to server. If None, do not use authentication.
    api_version : str or int (default: latest)
        What version of the api to use, 0: Legacy client (i.e www.dynamicannotationframework.com) 
        2: new api version, (i.e. minniev1.microns-daf.com)
        'latest': default to the most recent (current 2)

    Returns
    -------
    ClientBaseWithDatastack
        List of datastack names for available datastacks on the annotation engine
    """

    if auth_client is None:
        auth_client = AuthClient()

    auth_header = auth_client.request_header
    endpoints, api_version = _api_endpoints(api_version, SERVER_KEY, server_address,
                                            annotation_common, annotation_api_versions, auth_header)

    AnnoClient = client_mapping[api_version]
    if api_version > 1:
        return AnnoClient(server_address, auth_header, api_version,
                          endpoints, SERVER_KEY, aligned_volume_name)
    else:
        return AnnoClient(server_address, auth_header, api_version,
                          endpoints, SERVER_KEY, dataset_name)


class AnnotationClientLegacy(ClientBaseWithDataset):
    def __init__(self, server_address, auth_header, api_version, endpoints, server_name, dataset_name):
        super(AnnotationClientLegacy, self).__init__(server_address,
                                                     auth_header, api_version, endpoints,
                                                     server_name, dataset_name)

    def get_datasets(self):
        """ Gets a list of datasets

        Returns
        -------
        list
            List of dataset names for available datasets on the annotation engine
        """
        url = self._endpoints["datasets"].format_map(self.default_url_mapping)
        response = self.session.get(url)
        return handle_response(response)

    def get_tables(self, dataset_name=None):
        """ Gets a list of table names for a dataset

        Parameters
        ----------
        dataset_name : str or None, optional
            Name of the dataset, by default None. If None, uses the one specified in the client.

        Returns
        -------
        list
            List of table names
        """
        if dataset_name is None:
            dataset_name = self.dataset_name
        endpoint_mapping = self.default_url_mapping
        endpoint_mapping["dataset_name"] = dataset_name
        url = self._endpoints["table_names"].format_map(endpoint_mapping)

        response = self.session.get(url)
        return handle_response(response)

    def create_table(self, table_name, schema_name, dataset_name=None):
        """ Creates a new data table based on an existing schema

        Parameters
        ----------
        table_name: str
            Name of the new table. Cannot be the same as an existing table
        schema_name: str
            Name of the schema for the new table.
        dataset_name: str or None, optional,
            Name of the dataset. If None, uses the one specified in the client.

        Returns
        -------
        json
            Response JSON

        """
        if dataset_name is None:
            dataset_name = self.dataset_name

        endpoint_mapping = self.default_url_mapping
        endpoint_mapping["dataset_name"] = dataset_name
        url = self._endpoints["table_names"].format_map(endpoint_mapping)
        data = {"schema_name": schema_name, "table_name": table_name}

        response = requests.post(url, json=data)
        return handle_response(response)

    def get_annotation(self, table_name, annotation_id, dataset_name=None):
        """ Retrieve a single annotation by id and table name.

        Parameters
        ----------
        table_name : str
            Name of the table
        annotation_id : int
            ID number of the annotation to retreive (starting from 1)
        dataset_name : str or None, optional
            Name of the dataset. If None, uses the one specified in the client.

        Returns
        -------
        dict
            Annotation data
        """
        if dataset_name is None:
            dataset_name = self.dataset_name

        endpoint_mapping = self.default_url_mapping
        endpoint_mapping["dataset_name"] = dataset_name
        endpoint_mapping["table_name"] = table_name
        endpoint_mapping["annotation_id"] = annotation_id

        url = self._endpoints["existing_annotation"].format_map(
            endpoint_mapping)
        response = self.session.get(url)
        return handle_response(response)

    def post_annotation(self, table_name, data, dataset_name=None):
        """ Post one or more new annotations to a table in the AnnotationEngine

        Parameters
        ----------
        table_name : str
            Name of the table where annotations will be added
        data : dict or list,
            A list of (or a single) dict of schematized annotation data matching the target table.
        dataset_name : str or None, optional
            Name of the dataset. If None, uses the one specified in the client.

        Returns
        -------
        json
            Response JSON

        """
        if dataset_name is None:
            dataset_name = self.dataset_name
        if isinstance(data, dict):
            data = [data]

        endpoint_mapping = self.default_url_mapping
        endpoint_mapping["dataset_name"] = dataset_name
        endpoint_mapping["table_name"] = table_name

        url = self._endpoints["new_annotation"].format_map(endpoint_mapping)

        response = self.session.post(url, data=json.dumps(data, cls=AEEncoder),
                                     headers={'Content-Type': 'application/json'})
        return handle_response(response)


class AnnotationClientV2(ClientBase):
    def __init__(self, server_address, auth_header, api_version,
                 endpoints, server_name, aligned_volume_name):
        super(AnnotationClientV2, self).__init__(server_address,
                                                 auth_header, api_version, endpoints, server_name)

        self._aligned_volume_name = aligned_volume_name

    @property
    def aligned_volume_name(self):
        return self._aligned_volume_name

    def get_tables(self, aligned_volume_name=None):
        """ Gets a list of table names for a aligned_volume_name

        Parameters
        ----------
        aligned_volume_name : str or None, optional
            Name of the aligned_volume, by default None.
            If None, uses the one specified in the client.
            Will be set correctly if you are using the framework_client

        Returns
        -------
        list
            List of table names
        """
        if aligned_volume_name is None:
            aligned_volume_name = self.aligned_volume_name
        endpoint_mapping = self.default_url_mapping
        endpoint_mapping["aligned_volume_name"] = aligned_volume_name
        url = self._endpoints["tables"].format_map(endpoint_mapping)

        response = self.session.get(url)
        return handle_response(response)

    def get_annotation_count(self, table_name: str, aligned_volume_name=None):
        """ Get number of annotations in a table

        Parameters
        ----------
        table_name (str): 
            name of table to mark for deletion
        aligned_volume_name: str or None, optional,
            Name of the aligned_volume. If None, uses the one specified in the client.


        Returns
        -------
        int
            number of annotations
        """
        if aligned_volume_name is None:
            aligned_volume_name = self.aligned_volume_name

        endpoint_mapping = self.default_url_mapping
        endpoint_mapping["aligned_volume_name"] = aligned_volume_name
        endpoint_mapping["table_name"] = table_name

        url = self._endpoints["table_count"].format_map(endpoint_mapping)

        response = self.session.get(url)
        return handle_response(response)

    def get_table_metadata(self, table_name: str, aligned_volume_name=None):
        """ Get metadata about a table

        Parameters
        ----------
        table_name (str): 
            name of table to mark for deletion
        aligned_volume_name: str or None, optional,
            Name of the aligned_volume. If None, uses the one specified in the client.


        Returns
        -------
        json
            metadata about table
        """
        if aligned_volume_name is None:
            aligned_volume_name = self.aligned_volume_name

        endpoint_mapping = self.default_url_mapping
        endpoint_mapping["aligned_volume_name"] = aligned_volume_name
        endpoint_mapping["table_name"] = table_name

        url = self._endpoints["table_info"].format_map(endpoint_mapping)

        response = self.session.get(url)
        return handle_response(response)

    def delete_table(self, table_name: str, aligned_volume_name=None):
        """ Marks a table for deletion
        requires super admin priviledges

        Parameters
        ----------
        table_name (str): 
            name of table to mark for deletion
        aligned_volume_name: str or None, optional,
            Name of the aligned_volume. If None, uses the one specified in the client.


        Returns
        -------
        json
            Response JSON
        """
        if aligned_volume_name is None:
            aligned_volume_name = self.aligned_volume_name

        endpoint_mapping = self.default_url_mapping
        endpoint_mapping["aligned_volume_name"] = aligned_volume_name
        endpoint_mapping["table_name"] = table_name

        url = self._endpoints["table_info"].format_map(endpoint_mapping)

        response = self.session.delete(url)
        return handle_response(response)

    def create_table(self, table_name, schema_name,
                     description, reference_table=None,
                     flat_segmentation_source=None,
                     user_id=None,
                     aligned_volume_name=None):
        """ Creates a new data table based on an existing schema

        Parameters
        ----------
        table_name: str
            Name of the new table. Cannot be the same as an existing table
        schema_name: str
            Name of the schema for the new table.
        descrption: str
            Human readable description for what is in the table.
            Should include information about who generated the table
            What data it covers, and how it should be interpreted.
            And who should you talk to if you want to use it.
            An Example:
            a manual synapse table to detect chandelier synapses
            on 81 PyC cells with complete AISs 
            [created by Agnes - agnesb@alleninstitute.org, uploaded by Forrest]
        reference_table: str or None
            If the schema you are using is a reference schema
            Meaning it is an annotation of another annotation.
            Then you need to specify what table those annotations are in.
        flat_segmentation_source: str or None
            the source to a flat segmentation that corresponds to this table
            i.e. precomputed:\\gs:\\mybucket\this_tables_annotation
        user_id: int
            If you are uploading this schema on someone else's behalf
            and you want to link this table with their ID, you can specify it here
            Otherwise, the table will be created with your userID in the user_id column.
        aligned_volume_name: str or None, optional,
            Name of the aligned_volume. If None, uses the one specified in the client.

        Returns
        -------
        json
            Response JSON

        """
        if aligned_volume_name is None:
            aligned_volume_name = self.aligned_volume_name

        endpoint_mapping = self.default_url_mapping
        endpoint_mapping["aligned_volume_name"] = aligned_volume_name

        url = self._endpoints["tables"].format_map(endpoint_mapping)
        metadata = {'description': description}
        if user_id is not None:
            metadata['user_id'] = user_id
        if reference_table is not None:
            metadata['reference_table'] = reference_table
        if flat_segmentation_source is not None:
            metadata['flat_segmentation_source'] = flat_segmentation_source
        data = {"schema_type": schema_name,
                "table_name": table_name,
                "metadata": metadata}

        response = self.session.post(url, json=data)
        return handle_response(response)

    def get_annotation(self, table_name, annotation_ids, aligned_volume_name=None):
        """ Retrieve an annotation or annotations by id(s) and table name.

        Parameters
        ----------
        table_name : str
            Name of the table
        annotation_ids : int or iterable
            ID or IDS of the annotation to retreive
        aligned_volume_name : str or None, optional
            Name of the aligned_volume. If None, uses the one specified in the client.

        Returns
        -------
        list
            Annotation data
        """
        if aligned_volume_name is None:
            aligned_volume_name = self.aligned_volume_name

        endpoint_mapping = self.default_url_mapping
        endpoint_mapping["aligned_volume_name"] = aligned_volume_name
        endpoint_mapping["table_name"] = table_name
        url = self._endpoints["annotations"].format_map(endpoint_mapping)
        try:
            iter(annotation_ids)
        except TypeError:
            annotation_ids = [annotation_ids]

        params = {
            'annotation_ids': ",".join([str(a) for a in annotation_ids])
        }
        response = self.session.get(url, params=params)
        return handle_response(response)

    def post_annotation(self, table_name, data, aligned_volume_name=None):
        """ Post one or more new annotations to a table in the AnnotationEngine

        Parameters
        ----------
        table_name : str
            Name of the table where annotations will be added
        data : dict or list,
            A list of (or a single) dict of schematized annotation data matching the target table.
        aligned_volume_name : str or None, optional
            Name of the aligned_volume. If None, uses the one specified in the client.

        Returns
        -------
        json
            Response JSON

        """
        if aligned_volume_name is None:
            aligned_volume_name = self.aligned_volume_name
        if isinstance(data, dict):
            data = [data]

        endpoint_mapping = self.default_url_mapping
        endpoint_mapping["aligned_volume_name"] = aligned_volume_name
        endpoint_mapping["table_name"] = table_name
        url = self._endpoints["annotations"].format_map(endpoint_mapping)

        try:
            iter(data)
        except TypeError:
            annotation_ids = [data]

        data = {
            "annotations": data
        }

        response = self.session.post(url, data=json.dumps(data, cls=AEEncoder),
                                     headers={'Content-Type': 'application/json'})
        return handle_response(response)

    def update_annotation(self, table_name, data, aligned_volume_name=None):
        """Update one or more new annotations to a table in the AnnotationEngine
        Note update is implemented by deleting the old annotation
        and inserting a new annotation, which will receive a new ID.

        Parameters
        ----------
        table_name : str
            Name of the table where annotations will be added
        data : dict or list,
            A list of (or a single) dict of schematized annotation data matching the target table.
            each dict must contain an "id" field which is the ID of the annotation to update
        aligned_volume_name : str or None, optional
            Name of the aligned_volume. If None, uses the one specified in the client.

        Returns
        -------
        json
            Response JSON: a list of new annotation IDs.

        """
        if aligned_volume_name is None:
            aligned_volume_name = self.aligned_volume_name
        if isinstance(data, dict):
            data = [data]

        endpoint_mapping = self.default_url_mapping
        endpoint_mapping["aligned_volume_name"] = aligned_volume_name
        endpoint_mapping["table_name"] = table_name
        url = self._endpoints["annotations"].format_map(endpoint_mapping)

        try:
            iter(data)
        except TypeError:
            annotation_ids = [data]

        data = {
            "annotations": data
        }

        response = self.session.put(url, json=data)
        return handle_response(response)

    def delete_annotation(self, table_name, annotation_ids, aligned_volume_name=None):
        """Update one or more new annotations to a table in the AnnotationEngine
        Note update is implemented by deleting the old annotation
        and inserting a new annotation, which will receive a new ID.

        Parameters
        ----------
        table_name : str
            Name of the table where annotations will be added
        data : dict or list,
            A list of (or a single) dict of schematized annotation data matching the target table.
            each dict must contain an "id" field which is the ID of the annotation to update
        aligned_volume_name : str or None, optional
            Name of the aligned_volume. If None, uses the one specified in the client.

        Returns
        -------
        json
            Response JSON: a list of new annotation IDs.

        """
        if aligned_volume_name is None:
            aligned_volume_name = self.aligned_volume_name

        endpoint_mapping = self.default_url_mapping
        endpoint_mapping["aligned_volume_name"] = aligned_volume_name
        endpoint_mapping["table_name"] = table_name
        url = self._endpoints["annotations"].format_map(endpoint_mapping)

        try:
            iter(annotation_ids)
        except TypeError:
            annotation_ids = [annotation_ids]

        data = {
            "annotation_ids": annotation_ids
        }

        response = self.session.delete(url, data=json.dumps(data, cls=AEEncoder),
                                       headers={'Content-Type': 'application/json'})
        return handle_response(response)


client_mapping = {0: AnnotationClientLegacy,
                  2: AnnotationClientV2,
                  'latest': AnnotationClientV2}



from setuptools import setup
import re
import os
import codecs

here = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    with codecs.open(os.path.join(here, *parts), 'r') as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


with open('requirements.txt', 'r') as f:
    required = f.read().splitlines()

setup(
    version=find_version("annotationframeworkclient", "__init__.py"),
    name='annotationframeworkclient',
    description='a service for interacting with the AnnotationFramework',
    author='Sven Dorkenwald',
    author_email='svenmd@princeton.edu',
    url='https://github.com/seung-lab/AnnotationFrameworkClient',
    packages=['annotationframeworkclient'],
    include_package_data=True,
    install_requires=required,
    setup_requires=['pytest-runner'])



from .annotationengine import AnnotationClient
from .auth import AuthClient, default_token_file
from .chunkedgraph import ChunkedGraphClient
from .emannotationschemas import SchemaClient
from .infoservice import InfoServiceClient
from .jsonservice import JSONService
from .materializationengine import MaterializationClient
from .endpoints import default_global_server_address


class GlobalClientError(Exception):
    pass


class FrameworkClient(object):
    def __new__(
        cls,
        datastack_name=None,
        server_address=None,
        auth_token_file=None,
        auth_token_key=None,
        auth_token=None,
        global_only=False,
    ):
        if global_only or datastack_name is None:
            return FrameworkClientGlobal(
                server_address=server_address,
                auth_token_file=auth_token_file,
                auth_token_key=auth_token_key,
                auth_token=auth_token,
            )
        else:
            return FrameworkClientFull(
                datastack_name=datastack_name,
                server_address=server_address,
                auth_token_file=auth_token_file,
                auth_token_key=auth_token_key,
                auth_token=auth_token,
            )


class FrameworkClientGlobal(object):
    """A manager for all clients sharing common datastack and authentication information.

    This client wraps all the other clients and keeps track of the things that need to be consistent across them.
    To instantiate a client:
    client = FrameworkClient(datastack_name='my_datastack', server_address='www.myserver.com',
                             auth_token_file='~/.mysecrets/secrets.json')

    Then
    * client.info is an InfoService client (see infoservice.InfoServiceClient)
    * client.auth handles authentication
    * client.state is a neuroglancer state client (see jsonservice.JSONService)
    * client.schema is an EM Annotation Schemas client (see emannotationschemas.SchemaClient)

    All subclients are loaded lazily and share the same datastack name, server address, and auth tokens (where used).

    Parameters
    ----------
    server_address : str or None
        URL of the framework server. If None, chooses the default server www.dynamicannotationframework.com.
        Optional, defaults to None.
    auth_token_file : str or None
        Path to a json file containing the auth token. If None, uses the default location. See Auth client documentation.
        Optional, defaults to None.
    auth_token_key : str
        Dictionary key for the token in the the JSON file.
        Optional, default is 'token'.
    auth_token : str or None
        Direct entry of an auth token. If None, uses the file arguments to find the token.
        Optional, default is None.
    """

    def __init__(
        self,
        server_address=None,
        auth_token_file=None,
        auth_token_key=None,
        auth_token=None,
    ):
        if server_address is None:
            server_address = default_global_server_address
        self._server_address = server_address
        self._auth_config = {}
        self.change_auth(auth_token_file=auth_token_file,
                         auth_token_key=auth_token_key,
                         auth_token=auth_token)

    def change_auth(self, auth_token_file=None, auth_token_key=None, auth_token=None):
        """Change the authentication token and reset services.

        Parameters
        ----------
        auth_token_file : str, optional
            New auth token json file path, by default None, which defaults to the existing state.
        auth_token_key : str, optional
            New dictionary key under which the token is stored in the json file, by default None,
            which defaults to the existing state.
        auth_token : str, optional
            Direct entry of a new token, by default None.
        """
        if auth_token_file is None:
            auth_token_file = self._auth_config.get('auth_token_file', None)
        if auth_token_key is None:
            auth_token_key = self._auth_config.get('auth_token_key', None)

        self._auth_config = {
            'token_file': auth_token_file,
            'token_key': auth_token_key,
            'token': auth_token,
            'server_address': self._server_address,
        }
        self._reset_services()

    def _reset_services(self):
        self._auth = None
        self._info = None
        self._state = None
        self._schema = None

    @property
    def server_address(self):
        return self._server_address

    @property
    def auth(self):
        if self._auth is None:
            self._auth = AuthClient(**self._auth_config)
        return self._auth

    @property
    def info(self) -> InfoServiceClient:
        if self._info is None:
            self._info = InfoServiceClient(
                server_address=self.server_address,
                datastack_name=self.datastack_name,
                auth_client=self.auth,
            )
        return self._info

    @property
    def state(self):
        if self._state is None:
            self._state = JSONService(
                server_address=self.server_address, auth_client=self.auth
            )
        return self._state

    @property
    def schema(self):
        if self._schema is None:
            self._schema = SchemaClient(
                server_address=self.server_address, auth_client=self.auth
            )
        return self._schema

    def _no_local_functionality(self):
        raise GlobalClientError(
            "Client in global-only mode because no datastack was set."
        )

    @property
    def annotation(self):
        self._no_local_functionality()

    @property
    def chunkedgraph(self):
        self._no_local_functionality()

    @property
    def datastack_name(self):
        return None


class FrameworkClientFull(FrameworkClientGlobal):
    """A manager for all clients sharing common datastack and authentication information.

    This client wraps all the other clients and keeps track of the things that need to be consistent across them.
    To instantiate a client:
    client = FrameworkClient(datastack_name='my_datastack', server_address='www.myserver.com',
                             auth_token_file='~/.mysecrets/secrets.json')

    Then
    * client.info is an InfoService client (see infoservice.InfoServiceClient)
    * client.state is a neuroglancer state client (see jsonservice.JSONService)
    * client.schema is an EM Annotation Schemas client (see emannotationschemas.SchemaClient)
    * client.chunkedgraph is a Chunkedgraph client (see chunkedgraph.ChunkedGraphClient)
    * client.annotation is an Annotation DB client (see annotationengine.AnnotationClient)

    All subclients are loaded lazily and share the same datastack name, server address, and auth tokens where used.

    Parameters
    ----------
    datastack_name : str, optional
        Datastack name for the services. Almost all services need this and will not work if it is not passed.
    server_address : str or None
        URL of the framework server. If None, chooses the default server www.dynamicannotationframework.com.
        Optional, defaults to None.
    auth_token_file : str or None
        Path to a json file containing the auth token. If None, uses the default location. See Auth client documentation.
        Optional, defaults to None.
    auth_token_key : str
        Dictionary key for the token in the the JSON file.
        Optional, default is 'token'.
    auth_token : str or None
        Direct entry of an auth token. If None, uses the file arguments to find the token.
        Optional, default is None.
    """

    def __init__(
        self,
        datastack_name=None,
        server_address=None,
        auth_token_file=default_token_file,
        auth_token_key="token",
        auth_token=None,
    ):
        super(FrameworkClientFull, self).__init__(
            server_address=server_address,
            auth_token_file=auth_token_file,
            auth_token_key=auth_token_key,
            auth_token=auth_token,
        )

        self._datastack_name = datastack_name

        self._chunkedgraph = None
        self._annotation = None
        self._materialize = None

        self.local_server = self.info.local_server()
        av_info = self.info.get_aligned_volume_info()
        self._aligned_volume_name = av_info["name"]

    def _reset_services(self):
        self._auth = None
        self._info = None
        self._state = None
        self._schema = None
        self._chunkedgraph = None
        self._annotation = None
        self._materialize = None

    @property
    def datastack_name(self):
        return self._datastack_name

    @property
    def chunkedgraph(self):
        seg_source = self.info.segmentation_source()
        table_name = seg_source.split("/")[-1]

        if self._chunkedgraph is None:
            self._chunkedgraph = ChunkedGraphClient(
                table_name=table_name,
                server_address=self.local_server,
                auth_client=self.auth,
            )
        return self._chunkedgraph

    @property
    def annotation(self):
        if self._annotation is None:
            self._annotation = AnnotationClient(
                server_address=self.local_server,
                aligned_volume_name=self._aligned_volume_name,
                auth_client=self.auth,
            )
        return self._annotation

    @property
    def materialize(self):
        if self._materialize is None:
            self._materialize = MaterializationClient(server_address=self.local_server,
                                                      auth_client=self.auth,
                                                      datastack_name=self._datastack_name)
        return self._materialize
        
    @property
    def state(self):
        if self._state is None:
            self._state = JSONService(
                server_address=self.server_address, auth_client=self.auth, ngl_url=self.info.viewer_site(),
            )
        return self._state
